from tqsdk import TqApi, TqAuth

'''
如果买入看涨期权构建多头期货的价格小于卖出期货价格
存在套利机会则发出双边挂单
'''
api = TqApi(auth=TqAuth("快期账户", "账号密码"))

# 获取行权价为2950的MA109看涨期权的quote数据
quote_option = api.get_quote('CZCE.MA109C2950')

# 获取期权对应标的期货，即MA109的quote数据
quote = api.get_quote(quote_option.underlying_symbol)

# 套利机会尝试次数
times = 0

while True:
    api.wait_update()
    # 以对手价来计算买入看涨期权然后行权后的期货成本价格
    option_buy = quote_option.strike_price + quote_option.ask_price1
    # 判断当期货的买入1档，即卖出期货价格大于买入看涨期权的期货成本价格，形成套利空间时进行限价下单
    if quote.bid_price1 > option_buy and times == 0:
        times += 1
        # 以现在卖1档价格买入看涨期权
        order_opiton = api.insert_order('CZCE.MA109C2950', "BUY", "OPEN", 1, quote_option.ask_price1)
        # 以现在买1档的价格卖出期货
        order_future = api.insert_order(quote.underlying_symbol, "SELL", "OPEN", 1, quote.bid_price1)
        print("存在期货，期权套利空间尝试买入")



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'limin'

from pandas import Series

"""
tqsdk.tafunc 模块包含了一批用于技术指标计算的函数
(函数基本保持 参数为pandas.Series类型则返回值为pandas.Series类型)
"""

import datetime
import math
from typing import Union

import numpy as np
import pandas as pd
from scipy import stats

from tqsdk.datetime import _get_period_timestamp, _str_to_timestamp_nano, _datetime_to_timestamp_nano


def ref(series, n):
    """
    简单移动: 求series序列位移n个周期的结果

        注意: 当n为0, 函数返回原序列; 当n为有效值但当前的series序列元素个数不足 n + 1 个, 函数返回 NaN 序列

    Args:
        series (pandas.Series): 数据序列

        n (int): 位移周期

    Returns:
        pandas.Series: 位移后的序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        pre_close = tafunc.ref(klines.close, 1)  # 将收盘价序列右移一位, 得到昨收盘序列
        change = klines.close - pre_close        # 收盘价序列 - 昨收盘序列, 得到涨跌序列
        print(list(change))
    """
    m = series.shift(n)
    return m


def std(series, n):
    """
    标准差: 求series序列每n个周期的标准差

        注意: n为0的情况下, 或当n为有效值但当前的series序列元素个数不足n个, 函数返回 NaN 序列

    Args:
        series (pandas.Series): 数据序列

        n (int): 标准差的周期

    Returns:
        pandas.Series: 标准差序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        std = tafunc.std(klines.close, 5)  # 收盘价序列每5项计算一个标准差
        print(list(std))
    """
    m = series.rolling(n).std()
    return m


def ma(series, n):
    """
    简单移动平均线: 求series序列n周期的简单移动平均

        计算公式:
        ma(x, 5) = (x(1) + x(2) + x(3) + x(4) + x(5)) / 5

        注意:
        1. 简单移动平均线将设定周期内的值取平均值, 其中各元素的权重都相等
        2. n为0的情况下, 或当n为有效值但当前的series序列元素个数不足n个, 函数返回 NaN 序列

    Args:
        series (pandas.Series): 数据序列

        n (int): 周期

    Returns:
        pandas.Series: 简单移动平均值序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        ma = tafunc.ma(klines.close, 5)
        print(list(ma))
    """
    ma_data = series.rolling(n).mean()
    return ma_data


def sma(series, n, m):
    """
    扩展指数加权移动平均: 求series序列n周期的扩展指数加权移动平均
    
        计算公式:
        sma(x, n, m) = sma(x, n, m).shift(1) * (n - m) / n + x(n) * m / n
        
        注意: n必须大于m

    Args:
        series (pandas.Series): 数据序列
        
        n (int): 周期
        
        m (int): 权重

    Returns:
        pandas.Series: 扩展指数加权移动平均序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        sma = tafunc.sma(klines.close, 5, 2)  # 收盘价序列每5项计算一个扩展指数加权移动平均值
        print(list(sma))
    """
    sma_data = series.ewm(alpha=m / n, adjust=False).mean()
    return sma_data


def ema(series, n):
    """
    指数加权移动平均线: 求series序列n周期的指数加权移动平均

        计算公式:
            ema(x, n) = 2 * x / (n + 1) + (n - 1) * ema(x, n).shift(1) / (n + 1)

        注意:
            1. n 需大于等于1
            2. 对距离当前较近的k线赋予了较大的权重

    Args:
        series (pandas.Series): 数据序列

        n (int): 周期

    Returns:
        pandas.Series: 指数加权移动平均线序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        ema = tafunc.ema(klines.close, 5)
        print(list(ema))
    """
    ema_data = series.ewm(span=n, adjust=False).mean()
    return ema_data


def ema2(series, n):
    """
    线性加权移动平均: 求series值的n周期线性加权移动平均 (也称WMA)

        计算公式:
            ema2(x, n) = [n * x(0) + (n - 1) * x(1) + (x - 2) * x(2) + ... + 1 * x(n - 1)] / [n + (n - 1) + (n - 2) + ... + 1]

        注意: 当n为有效值但当前的series序列元素个数不足n个, 函数返回 NaN 序列

    Args:
        series (pandas.Series): 数据序列

        n (int): 周期

    Returns:
        pandas.Series: 线性加权移动平均线序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        ema2 = tafunc.ema2(klines.close, 5)  # 求收盘价在5个周期的线性加权移动平均值
        print(list(ema2))
    """
    weights = list(i for i in range(1, n + 1))  # 对应的权值列表

    def average(elements):
        return np.average(elements, weights=weights)

    ema2 = series.rolling(window=n).apply(average, raw=True)
    return ema2


def crossup(a, b):
    """
    向上穿越: 表当a从下方向上穿过b, 成立返回1, 否则返回0

    Args:
        a (pandas.Series): 数据序列1

        b (pandas.Series): 数据序列2

    Returns:
        pandas.Series: 上穿标志序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        crossup = tafunc.crossup(tafunc.ma(klines.close, 5), tafunc.ma(klines.close, 10))
        print(list(crossup))
    """
    crossup_data = pd.Series(np.where((a > b) & (a.shift(1) <= b.shift(1)), 1, 0))
    return crossup_data


def crossdown(a, b):
    """
    向下穿越: 表示当a从上方向下穿b，成立返回1, 否则返回0

    Args:
        a (pandas.Series): 数据序列1

        b (pandas.Series): 数据序列2

    Returns:
        pandas.Series: 下穿标志序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc
        
        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        crossdown = tafunc.crossdown(tafunc.ma(klines.close, 5), tafunc.ma(klines.close, 10))
        print(list(crossdown))
    """
    crossdown_data = pd.Series(np.where((a < b) & (a.shift(1) >= b.shift(1)), 1, 0))
    return crossdown_data


def count(cond, n):
    """
    统计n周期中满足cond条件的个数

        注意: 如果n为0, 则从第一个有效值开始统计

    Args:
        cond (array_like): 条件

        n (int): 周期

    Returns:
        pandas.Series: 统计值序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        # 统计从申请到的行情数据以来到当前这段时间内, 5周期均线上穿10周期均线的次数:
        count = tafunc.count(tafunc.crossup(tafunc.ma(klines.close, 5), tafunc.ma(klines.close, 10)), 0)
        print(list(count))
    """
    if n == 0:  # 从第一个有效值开始统计
        count_data = pd.Series(np.where(cond, 1, 0).cumsum())
    else:  # 统计每个n周期
        count_data = pd.Series(pd.Series(np.where(cond, 1, 0)).rolling(n).sum())
    return count_data


def trma(series, n):
    """
    三角移动平均: 求series的n周期三角移动平均值

        计算方法:
            三角移动平均线公式, 是采用算数移动平均, 并且对第一个移动平均线再一次应用算数移动平均

        注意: n为0的情况下, 或当n为有效值但当前的series序列元素个数不足n个, 函数返回 NaN 序列

    Args:
        series (pandas.Series): 数据序列

        n (int): 周期

    Returns:
        pandas.Series: 三角移动平均值序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        trma = tafunc.trma(klines.close, 10)
        print(list(trma))
    """
    if n % 2 == 0:
        n1 = int(n / 2)
        n2 = int(n / 2 + 1)
    else:
        n1 = n2 = int((n + 1) / 2)
    ma_half = ma(series, n1)
    trma_data = ma(ma_half, n2)
    return trma_data


def harmean(series, n):
    """
    调和平均值: 求series在n个周期内的调和平均值

        计算方法:
            harmean(x, 5) = 1 / [(1 / x(1) + 1 / x(2) + 1 / x(3) + 1 / x(4) + 1 / x(5)) / 5]

        注意:
        1. 调和平均值与倒数的简单平均值互为倒数
        2. 当n为0, 或当n为有效值但当前的series序列元素个数不足n个, 函数返回 NaN 序列

    Args:
        series (pandas.Series): 数据序列

        n (int): 周期

    Returns:
        pandas.Series: 调和平均值序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        harmean = tafunc.harmean(klines.close, 5)  # 求5周期收盘价的调和平均值
        print(list(harmean))
    """
    harmean_data = n / ((1 / series).rolling(n).sum())
    return harmean_data


def numpow(series, n, m):
    """
    自然数幂方和

        计算方法:
            numpow(x, n, m) = n ^ m * x + (n - 1) ^ m * x.shift(1) + (n - 2) ^ m * x.shift(2) + ... + 2 ^ m * x.shift(n - 2) + 1 ^ m * x.shift(n - 1)

        注意: 当n为有效值但当前的series序列元素个数不足n个, 函数返回 NaN 序列

    Args:
        series (pandas.Series): 数据序列

        n (int): 自然数

        m (int): 实数

    Returns:
        pandas.Series: 幂方和序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        numpow = tafunc.numpow(klines.close, 5, 2)
        print(list(numpow))
    """
    numpow_data = sum((n - i) ** m * series.shift(i) for i in range(n))
    return numpow_data


def abs(series):
    """
    获取series的绝对值

        注意: 正数的绝对值是它本身, 负数的绝对值是它的相反数, 0的绝对值还是0

    Args:
        series (pandas.Series): 数据序列

    Returns:
        pandas.Series: 绝对值序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        abs = tafunc.abs(klines.close)
        print(list(abs))
    """
    abs_data = pd.Series(np.absolute(series))
    return abs_data


def min(series1, series2):
    """
    获取series1和series2中的最小值

    Args:
        series1 (pandas.Series): 数据序列1

        series2 (pandas.Series): 数据序列2

    Returns:
        pandas.Series: 最小值序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        min = tafunc.min(klines.close, klines.open)
        print(list(min))
    """
    min_data = np.minimum(series1, series2)
    return min_data


def max(series1, series2):
    """
    获取series1和series2中的最大值

    Args:
        series1 (pandas.Series): 数据序列1

        series2 (pandas.Series): 数据序列2

    Returns:
        pandas.Series: 最大值序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        max = tafunc.max(klines.close, klines.open)
        print(list(max))
    """
    max_data = np.maximum(series1, series2)
    return max_data


def median(series, n):
    """
    中位数: 求series在n个周期内居于中间的数值
    
        注意:
            1. 当n为有效值但当前的series序列元素个数不足n个, 函数返回 NaN 序列
            2. 对n个周期内所有series排序后, 若n为奇数, 则选择第(n + 1) / 2个为中位数, 若n为偶数, 则中位数是(n / 2)以及(n / 2 + 1)的平均数

    Args:
        series (pandas.Series): 数据序列

        n (int): 周期

    Returns:
        pandas.Series: 中位数序列

    Example::

        例1:
            # 假设最近3日的收盘价为2727, 2754, 2748, 那么当前 median(df["close"], 3) 的返回值是2748
            median3 = tafunc.median(df["close"], 3)

        例2:
            # 假设最近4日的开盘价为2752, 2743, 2730, 2728, 那么当前 median(df["open"], 4) 的返回值是2736.5
            median4 = tafunc.median(df["open"], 4)
    """
    median_data = series.rolling(n).median()
    return median_data


def exist(cond, n):
    """
    判断n个周期内, 是否有满足cond的条件, 若满足则值为1, 不满足为0

    Args:
        cond (array_like): 条件

        n (int): 周期

    Returns:
        pandas.Series: 判断结果序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        # 判断4个周期中是否存在收盘价大于前一个周期的最高价, 存在返回1, 不存在则返回0
        exist = tafunc.exist(klines.close > klines.high.shift(1), 4)
        print(list(exist))
    """
    exist_data = pd.Series(np.where(pd.Series(np.where(cond, 1, 0)).rolling(n).sum() > 0, 1, 0))
    return exist_data


def every(cond, n):
    """
    判断n个周期内, 是否一直满足cond条件, 若满足则值为1, 不满足为0

    Args:
        cond (array_like): 条件

        n (int): 周期

    Returns:
        pandas.Series: 判断结果序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        # 判断在4周期内, 3周期的简单移动平均是否一直大于5周期的简单移动平均
        every = tafunc.every(tafunc.ma(klines.close, 3) > tafunc.ma(klines.close, 5), 4)
        print(list(every))
    """
    every_data = pd.Series(np.where(pd.Series(np.where(cond, 1, 0)).rolling(n).sum() == n, 1, 0))
    return every_data


def hhv(series, n):
    """
    求series在n个周期内的最高值

        注意: n为0的情况下, 或当n为有效值但当前的series序列元素个数不足n个, 函数返回 NaN 序列

    Args:
        series (pandas.Series): 数据序列

        n (int): 周期

    Returns:
        pandas.Series: 最高值序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        hhv = tafunc.hhv(klines.high, 4)  # 求4个周期最高价的最大值, 即4周期高点(包含当前k线)
        print(list(hhv))
    """
    hhv_data = series.rolling(n).max()
    return hhv_data


def llv(series, n):
    """
    求在n个周期内的最小值

        注意: n为0的情况下, 或当n为有效值但当前的series序列元素个数不足n个, 函数返回 NaN 序列

    Args:
        series (pandas.Series): 数据序列

        n (int): 周期

    Returns:
        pandas.Series: 最小值序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        llv = tafunc.llv(klines.low, 5)  # 求5根k线最低点(包含当前k线)
        print(list(llv))
    """
    llv_data = series.rolling(n).min()
    return llv_data


def avedev(series, n):
    """
    平均绝对偏差: 求series在n周期内的平均绝对偏差

        算法:
            计算avedev(df["close"],3)在最近一根K线上的值:
            (abs(df["close"] - (df["close"] + df["close"].shift(1) + df["close"].shift(2)) / 3) + abs(
            df["close"].shift(1) - (df["close"] + df["close"].shift(1) + df["close"].shift(2)) / 3) + abs(
            df["close"].shift(2) - (df["close"] + df["close"].shift(1) + df["close"].shift(2)) / 3)) / 3

        注意: n为0的情况下, 或当n为有效值但当前的series序列元素个数不足n个, 函数返回 NaN 序列

    Args:
        series (pandas.Series): 数据序列

        n (int): 周期

    Returns:
        pandas.Series: 平均绝对偏差序列

    Example::

        from tqsdk import TqApi, TqAuth, TqSim, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("CFFEX.IF1908", 24 * 60 * 60)
        # 计算收盘价在5周期内的平均绝对偏差, 表示5个周期内每个周期的收盘价与5周期收盘价的平均值的差的绝对值的平均值, 判断收盘价与其均值的偏离程度:
        avedev = tafunc.avedev(klines.close, 5)
        print(list(avedev))
    """

    def mad(x):
        return np.fabs(x - x.mean()).mean()

    avedev_data = series.rolling(window=n).apply(mad, raw=True)
    return avedev_data


def _to_ns_timestamp(input_time):
    """
    辅助函数: 将传入的时间转换为int类型的纳秒级时间戳

    Args:
    input_time (str/ int/ float/ datetime.datetime): 需要转换的时间:
        * str: str 类型的时间，如Quote行情时间的datetime字段 (eg. 2019-10-14 14:26:01.000000)

        * int: int 类型纳秒级或秒级时间戳

        * float: float 类型纳秒级或秒级时间戳，如K线或tick的datetime字段 (eg. 1.57103449e+18)

        * datetime.datetime: datetime 模块中 datetime 类型

    Returns:
        int : int 类型纳秒级时间戳
    """

    if type(input_time) in {int, float, np.float64, np.float32, np.int64, np.int32}:  # 时间戳
        if input_time > 2 ** 32:  # 纳秒( 将 > 2*32数值归为纳秒级)
            return int(input_time)
        else:  # 秒
            return int(input_time * 1e9)
    elif isinstance(input_time, str):  # str 类型时间
        return _str_to_timestamp_nano(input_time)
    elif isinstance(input_time, datetime.datetime):  # datetime 类型时间
        return _datetime_to_timestamp_nano(input_time)
    else:
        raise TypeError("暂不支持此类型的转换")


def time_to_ns_timestamp(input_time):
    """
    将传入的时间转换为int类型的纳秒级时间戳

    Args:
        input_time (str/ int/ float/ datetime.datetime): 需要转换的时间:
            * str: str 类型的时间，如Quote行情时间的datetime字段 (eg. 2019-10-14 14:26:01.000000)

            * int: int 类型的纳秒级或秒级时间戳

            * float: float 类型的纳秒级或秒级时间戳，如K线或tick的datetime字段 (eg. 1.57103449e+18)

            * datetime.datetime: datetime 模块中的 datetime 类型时间

    Returns:
        int : int 类型的纳秒级时间戳

    Example::

        from tqsdk.tafunc import time_to_ns_timestamp
        print(time_to_ns_timestamp("2019-10-14 14:26:01.000000"))  # 将%Y-%m-%d %H:%M:%S.%f 格式的str类型转为纳秒时间戳
        print(time_to_ns_timestamp(1571103122))  # 将秒级转为纳秒时间戳
        print(time_to_ns_timestamp(datetime.datetime(2019, 10, 14, 14, 26, 1)))  # 将datetime.datetime时间转为纳秒时间戳
    """
    return _to_ns_timestamp(input_time)


def time_to_s_timestamp(input_time):
    """
    将传入的时间转换为int类型的秒级时间戳

    Args:
        input_time (str/ int/ float/ datetime.datetime): 需要转换的时间:
            * str: str 类型的时间，如Quote行情时间的datetime字段 (eg. 2019-10-14 14:26:01.000000)

            * int: int 类型的纳秒级或秒级时间戳

            * float: float 类型的纳秒级或秒级时间戳，如K线或tick的datetime字段 (eg. 1.57103449e+18)

            * datetime.datetime: datetime 模块中的 datetime 类型时间

    Returns:
        int : int类型的秒级时间戳

    Example::

        from tqsdk.tafunc import time_to_s_timestamp
        print(time_to_s_timestamp(1.57103449e+18))  # 将纳秒级时间戳转为秒级时间戳
        print(time_to_s_timestamp("2019-10-14 14:26:01.000000"))  # 将%Y-%m-%d %H:%M:%S.%f 格式的str类型时间转为秒级时间戳
        print(time_to_s_timestamp(datetime.datetime(2019, 10, 14, 14, 26, 1)))  # 将datetime.datetime时间转为秒时间戳
    """
    return int(_to_ns_timestamp(input_time) / 1e9)


def time_to_str(input_time):
    """
    将传入的时间转换为 %Y-%m-%d %H:%M:%S.%f 格式的 str 类型

    Args:
        input_time (int/ float/ datetime.datetime): 需要转换的时间:

            * int: int 类型的纳秒级或秒级时间戳

            * float: float 类型的纳秒级或秒级时间戳，如K线或tick的datetime字段 (eg. 1.57103449e+18)

            * datetime.datetime: datetime 模块中的 datetime 类型时间

    Returns:
        str : %Y-%m-%d %H:%M:%S.%f 格式的 str 类型时间

    Example::

        from tqsdk.tafunc import time_to_str
        print(time_to_str(1.57103449e+18))  # 将纳秒级时间戳转为%Y-%m-%d %H:%M:%S.%f 格式的str类型时间
        print(time_to_str(1571103122))  # 将秒级时间戳转为%Y-%m-%d %H:%M:%S.%f 格式的str类型时间
        print(time_to_str(datetime.datetime(2019, 10, 14, 14, 26, 1)))  # 将datetime.datetime时间转为%Y-%m-%d %H:%M:%S.%f 格式的str类型时间
    """
    # 转为秒级时间戳
    ts = _to_ns_timestamp(input_time) / 1e9
    # 转为 %Y-%m-%d %H:%M:%S.%f 格式的 str 类型时间
    dt = datetime.datetime.fromtimestamp(ts)
    dt = dt.strftime('%Y-%m-%d %H:%M:%S.%f')
    return dt


def time_to_datetime(input_time):
    """
    将传入的时间转换为 datetime.datetime 类型

    Args:
        input_time (int/ float/ str): 需要转换的时间:

            * int: int 类型的纳秒级或秒级时间戳

            * float: float 类型的纳秒级或秒级时间戳，如K线或tick的datetime字段 (eg. 1.57103449e+18)

            * str: str 类型的时间，如Quote行情时间的 datetime 字段 (eg. 2019-10-14 14:26:01.000000)

    Returns:
        datetime.datetime : datetime 模块中的 datetime 类型时间

    Example::

        from tqsdk.tafunc import time_to_datetime
        print(time_to_datetime(1.57103449e+18))  # 将纳秒级时间戳转为datetime.datetime时间
        print(time_to_datetime(1571103122))  # 将秒级时间戳转为datetime.datetime时间
        print(time_to_datetime("2019-10-14 14:26:01.000000"))  # 将%Y-%m-%d %H:%M:%S.%f 格式的str类型时间转为datetime.datetime时间
    """
    # 转为秒级时间戳
    ts = _to_ns_timestamp(input_time) / 1e9
    # 转为datetime.datetime类型
    dt = datetime.datetime.fromtimestamp(ts)
    return dt


def barlast(cond):
    """
    返回一个序列，其中每个值表示从上一次条件成立到当前的周期数

    (注： 如果从cond序列第一个值到某个位置之间没有True，则此位置的返回值为 -1； 条件成立的位置上的返回值为0)


    Args:
        cond (pandas.Series): 条件序列(序列中的值需为 True 或 False)

    Returns:
        pandas.Series : 周期数序列（其长度和 cond 相同；最后一个值即为最后一次条件成立到最新一个数据的周期数）

    Example::

        from tqsdk import TqApi, TqAuth
        from tqsdk.tafunc import barlast

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        klines = api.get_kline_serial("SHFE.cu1912", 60)
        # print(list(klines.close))
        # print(list(klines.open))
        # print(list(klines.close > klines.open))
        n = barlast(klines.close > klines.open)  # 获取周期数序列
        print(list(n))
        print(n.iloc[-1])  # 获取最后一根k线到上一次满足 "收盘价大于开盘价" 条件的k线的周期数
        api.close()

    """
    cond = cond.to_numpy()
    v = np.array(~cond, dtype=int)
    c = np.cumsum(v)
    x = c[cond]
    d = np.diff(np.concatenate(([0], x)))
    if len(d) == 0:  # 如果cond长度为0或无True
        return pd.Series([-1] * len(cond))
    v[cond] = -d
    r = np.cumsum(v)
    r[:x[0]] = -1
    return pd.Series(r)


def _get_t_series(series: pd.Series, dur: int, expire_datetime: int):
    t = pd.Series(pd.to_timedelta(expire_datetime - (series / 1e9 + dur), unit='s'))
    return (t.dt.days * 86400 + t.dt.seconds) / (360 * 86400)


def _get_d1(series: pd.Series, k: float, r: float, v: Union[float, pd.Series], t: Union[float, pd.Series]):
    return pd.Series(
        np.where((v <= 0) | (t <= 0), np.nan, (np.log(series / k) + (r + 0.5 * np.power(v, 2)) * t) / (v * np.sqrt(t))))


def _get_cdf(series: pd.Series):
    s = series.loc[series.notna()]
    return pd.concat([series.loc[series.isna()], pd.Series(stats.norm.cdf(s), index=s.index)], verify_integrity=True)


def _get_pdf(series: pd.Series):
    s = series.loc[series.notna()]
    return pd.concat([series.loc[series.isna()], pd.Series(stats.norm.pdf(s), index=s.index)], verify_integrity=True)


def _get_options_class(series: pd.Series, option_class: Union[str, pd.Series]):
    """
    根据价格序列 series，和指定的 option_class

    Args:
        option_class (str / Series[str]): CALL / PUT / Series(['CALL', 'CALL', 'CALL', 'PUT'])

    Returns:
        Series[int] :  长度和 series 一致，Series([1, 1, 1, 1]) / Series([-1, -1, -1, -1]) / Series([1, 1, 1, -1]), 对于无效的参数值为 Series([nan, nan, nan, nan])

    """
    if type(option_class) is str and option_class in ["CALL", "PUT"]:
        return Series([(1 if option_class == "CALL" else -1) for _ in range(series.size)])
    elif type(option_class) is Series and series.size == option_class.size:
        return option_class.map({'CALL': 1, 'PUT': -1})
    else:
        return Series([float('nan') for _ in range(series.size)])


def get_t(df, expire_datetime):
    """
    计算 K 线序列对应的年化到期时间，主要用于计算期权相关希腊指标时，需要得到计算出序列对应的年化到期时间

    Args:
        df (pandas.DataFrame): Dataframe 格式的 K 线序列

        expire_datetime (int): 到期日, 秒级时间戳

    Returns:
        pandas.Series : 返回的 df 对应的年化时间序列

    Example::

        from tqsdk import TqApi, TqAuth, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        quote = api.get_quote('SHFE.cu2006C45000')
        klines = api.get_kline_serial(['SHFE.cu2006C45000', 'SHFE.cu2006'], 24 * 60 * 60, 50)
        t = tafunc.get_t(klines, quote.expire_datetime)
        print(t)
        api.close()
    """
    return pd.Series(_get_t_series(df["datetime"], df["duration"], expire_datetime))


def get_his_volatility(df, quote):
    """
    计算某个合约的历史波动率

    Args:
        df (pandas.DataFrame): Dataframe 格式的 K 线序列

        quote (tqsdk.objs.Quote): df 序列对应合约对象

    Returns:
        float : 返回的 df 对应的历史波动率

    Example::

        from tqsdk import TqApi, TqAuth, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        quote = api.get_quote('SHFE.cu2006')
        klines = api.get_kline_serial('SHFE.cu2006', 24 * 60 * 60, 50)
        v = tafunc.get_his_volatility(klines, quote)
        print(v)
        api.close()
    """
    if quote and quote.instrument_id == df["symbol"][0]:
        trading_time = quote.trading_time
    else:
        trading_time = None
    return _get_volatility(df["close"], df["duration"], trading_time)


def _get_volatility(series: pd.Series, dur: Union[pd.Series, int] = 86400, trading_time: list = None) -> float:
    series_u = np.log(series.shift(1)[1:] / series[1:])
    series_u = series_u[~np.isnan(series_u)]
    if series_u.size < 2:  # 自由度小于2无法计算，返回一个默认值
        return float("nan")
    seconds_per_day = 24 * 60 * 60
    dur = dur[0] if isinstance(dur, pd.Series) else dur
    if dur < 24 * 60 * 60 and trading_time:
        periods = _get_period_timestamp(0, trading_time.get("day", []) + trading_time.get("night", []))
        seconds_per_day = sum([p[1] - p[0] for p in periods]) / 1e9
    return math.sqrt((250 * seconds_per_day / dur) * np.cov(series_u))


def get_bs_price(series, k, r, v, t, option_class):
    """
    计算期权 BS 模型理论价格

    Args:
        series (pandas.Series): 标的价格序列

        k (float): 期权行权价

        r (float): 无风险利率

        v (float / pandas.Series): 波动率

            * float: 对于 series 中每个元素都使用相同的 v 计算理论价

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 v 中对应的值计算理论价

        t (float / pandas.Series): 年化到期时间，例如：还有 100 天到期，则年化到期时间为 100/360

            * float: 对于 series 中每个元素都使用相同的 t 计算理论价

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 t 中对应的值计算理论价

        option_class (str / pandas.Series): 期权方向，必须是两者其一，否则返回的序列值全部为 nan

            * str: "CALL" 或者 "PUT"

            * pandas.Series: 其元素个数应该和 series 元素个数相同，每个元素的值为 "CALL" 或者 "PUT"

    Returns:
        pandas.Series: 返回该序列理论价

    Example::

        import pandas as pd
        from tqsdk import TqApi, TqAuth, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        quote = api.get_quote("SHFE.cu2006")
        ks = api.get_kline_serial("SHFE.cu2006", 24 * 60 * 60, 10)
        v = tafunc.get_his_volatility(ks, quote)  # 历史波动率

        option = api.get_quote("SHFE.cu2006C45000")
        klines = api.get_kline_serial(["SHFE.cu2006C45000", "SHFE.cu2006"], 24 * 60 * 60, 10)
        t = tafunc.get_t(klines, option.expire_datetime)
        bs_price = tafunc.get_bs_price(klines["close1"], 45000, 0.025, v, t, option.option_class)  # 理论价
        print(list(bs_price.round(2)))
        api.close()
    """
    o = _get_options_class(series, option_class=option_class)
    d1 = _get_d1(series, k, r, v, t)
    d2 = pd.Series(np.where(np.isnan(d1), np.nan, d1 - v * np.sqrt(t)))
    return pd.Series(
        np.where(np.isnan(d1), np.nan, o * (series * _get_cdf(o * d1) - k * np.exp(-r * t) * _get_cdf(o * d2))))


def get_delta(series, k, r, v, t, option_class, d1=None):
    """
    计算期权希腊指标 delta 值

    Args:
        series (pandas.Series): 标的价格序列

        k (float): 期权行权价

        r (float): 无风险利率

        v (float / pandas.Series): 波动率

            * float: 对于 series 中每个元素都使用相同的 v 计算理论价

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 v 中对应的值计算理论价

        t (float / pandas.Series): 年化到期时间，例如：还有 100 天到期，则年化到期时间为 100/360

            * float: 对于 series 中每个元素都使用相同的 t 计算理论价

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 t 中对应的值计算理论价

        option_class (str / pandas.Series): 期权方向，必须是两者其一，否则返回的序列值全部为 nan

            * str: "CALL" 或者 "PUT"

            * pandas.Series: 其元素个数应该和 series 元素个数相同，每个元素的值为 "CALL" 或者 "PUT"

        d1 (None | pandas.Series): [可选] 序列对应的 BS 公式中 b1 值

    Returns:
        pandas.Series: 该序列的 delta 值


    Example::

        import pandas as pd
        from tqsdk import TqApi, TqAuth, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        quote = api.get_quote("SHFE.cu2006")
        ks = api.get_kline_serial("SHFE.cu2006", 24 * 60 * 60, 10)
        v = tafunc.get_his_volatility(ks, quote)  # 历史波动率

        option = api.get_quote("SHFE.cu2006C45000")
        klines = api.get_kline_serial(["SHFE.cu2006C45000", "SHFE.cu2006"], 24 * 60 * 60, 10)
        t = tafunc.get_t(klines, option.expire_datetime)
        impv = tafunc.get_impv(klines["close1"], klines["close"], 45000, 0.025, v, t, "CALL")
        delta = tafunc.get_delta(klines["close1"], 45000, 0.025, v, t, "CALL")
        print("delta", list(delta))
        api.close()

    """
    o = _get_options_class(series, option_class=option_class)
    if d1 is None:
        d1 = _get_d1(series, k, r, v, t)
    return pd.Series(np.where(np.isnan(d1), np.nan, pd.Series(o * _get_cdf(o * d1))))


def get_gamma(series, k, r, v, t, d1=None):
    """
    计算期权希腊指标 gamma 值

    Args:
        series (pandas.Series): 标的价格序列

        k (float): 期权行权价

        r (float): 无风险利率

        v (float / pandas.Series): 波动率

            * float: 对于 series 中每个元素都使用相同的 v 计算理论价

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 v 中对应的值计算理论价

        t (float / pandas.Series): 年化到期时间，例如：还有 100 天到期，则年化到期时间为 100/360

            * float: 对于 series 中每个元素都使用相同的 t 计算理论价

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 t 中对应的值计算理论价

        d1 (None | pandas.Series): [可选] 序列对应的 BS 公式中 b1 值

    Returns:
        pandas.Series: 该序列的 gamma 值


    Example::

        import pandas as pd
        from tqsdk import TqApi, TqAuth, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        quote = api.get_quote("SHFE.cu2006")
        ks = api.get_kline_serial("SHFE.cu2006", 24 * 60 * 60, 10)
        v = tafunc.get_his_volatility(ks, quote)  # 历史波动率

        option = api.get_quote("SHFE.cu2006C45000")
        klines = api.get_kline_serial(["SHFE.cu2006C45000", "SHFE.cu2006"], 24 * 60 * 60, 10)
        t = tafunc.get_t(klines, option.expire_datetime)
        impv = tafunc.get_impv(klines["close1"], klines["close"], 45000, 0.025, v, t, "CALL")
        gamma = tafunc.get_gamma(klines["close1"], 45000, 0.025, v, t)
        print("gamma", list(gamma))
        api.close()

    """
    if d1 is None:
        d1 = _get_d1(series, k, r, v, t)
    return pd.Series(np.where(np.isnan(d1), np.nan, _get_pdf(d1) / (series * v * np.sqrt(t))))


def get_theta(series, k, r, v, t, option_class, d1=None):
    """
    计算期权希腊指标 theta 值

    Args:
        series (pandas.Series): 标的价格序列

        k (float): 期权行权价

        r (float): 无风险利率

        v (float / pandas.Series): 波动率

            * float: 对于 series 中每个元素都使用相同的 v 计算理论价

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 v 中对应的值计算理论价

        t (float / pandas.Series): 年化到期时间，例如：还有 100 天到期，则年化到期时间为 100/360

            * float: 对于 series 中每个元素都使用相同的 t 计算理论价

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 t 中对应的值计算理论价

        option_class (str / pandas.Series): 期权方向，必须是两者其一，否则返回的序列值全部为 nan

            * str: "CALL" 或者 "PUT"

            * pandas.Series: 其元素个数应该和 series 元素个数相同，每个元素的值为 "CALL" 或者 "PUT"

        d1 (None | pandas.Series): [可选] 序列对应的 BS 公式中 b1 值

    Returns:
        pandas.Series: 该序列的 theta 值


    Example::

        import pandas as pd
        from tqsdk import TqApi, TqAuth, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        quote = api.get_quote("SHFE.cu2006")
        ks = api.get_kline_serial("SHFE.cu2006", 24 * 60 * 60, 10)
        v = tafunc.get_his_volatility(ks, quote)  # 历史波动率

        option = api.get_quote("SHFE.cu2006C45000")
        klines = api.get_kline_serial(["SHFE.cu2006C45000", "SHFE.cu2006"], 24 * 60 * 60, 10)
        t = tafunc.get_t(klines, option.expire_datetime)
        impv = tafunc.get_impv(klines["close1"], klines["close"], 45000, 0.025, v, t, "CALL")
        theta = tafunc.get_theta(klines["close1"], 45000, 0.025, v, t, "CALL")
        print("theta", list(theta))
        api.close()

    """
    o = _get_options_class(series, option_class=option_class)
    if d1 is None:
        d1 = _get_d1(series, k, r, v, t)
    d2 = pd.Series(np.where(np.isnan(d1), np.nan, d1 - v * np.sqrt(t)))
    return pd.Series(np.where(np.isnan(d1), np.nan, pd.Series(
        -v * series * _get_pdf(d1) / (2 * np.sqrt(t)) - o * r * k * np.exp(-r * t) * _get_cdf(o * d2))))


def get_vega(series, k, r, v, t, d1=None):
    """
    计算期权希腊指标 vega 值

    Args:
        series (pandas.Series): 标的价格序列

        k (float): 期权行权价

        r (float): 无风险利率

        v (float / pandas.Series): 波动率

            * float: 对于 series 中每个元素都使用相同的 v 计算理论价

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 v 中对应的值计算理论价

        t (float / pandas.Series): 年化到期时间，例如：还有 100 天到期，则年化到期时间为 100/360

            * float: 对于 series 中每个元素都使用相同的 t 计算理论价

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 t 中对应的值计算理论价

        d1 (None | pandas.Series): [可选] 序列对应的 BS 公式中 b1 值

    Returns:
        pandas.Series: 该序列的 vega 值


    Example::

        import pandas as pd
        from tqsdk import TqApi, TqAuth, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        quote = api.get_quote("SHFE.cu2006")
        ks = api.get_kline_serial("SHFE.cu2006", 24 * 60 * 60, 10)
        v = tafunc.get_his_volatility(ks, quote)  # 历史波动率

        option = api.get_quote("SHFE.cu2006C45000")
        klines = api.get_kline_serial(["SHFE.cu2006C45000", "SHFE.cu2006"], 24 * 60 * 60, 10)
        t = tafunc.get_t(klines, option.expire_datetime)
        impv = tafunc.get_impv(klines["close1"], klines["close"], 45000, 0.025, v, t, "CALL")
        vega = tafunc.get_vega(klines["close1"], 45000, 0.025, v, t)
        print("vega", list(vega))
        api.close()

    """
    if d1 is None:
        d1 = _get_d1(series, k, r, v, t)
    return pd.Series(np.where(np.isnan(d1), np.nan, series * np.sqrt(t) * _get_pdf(d1)))


def get_rho(series, k, r, v, t, option_class, d1=None):
    """
    计算期权希腊指标 rho 值

    Args:
        series (pandas.Series): 标的价格序列

        k (float): 期权行权价

        r (float): 无风险利率

        v (float / pandas.Series): 波动率

            * float: 对于 series 中每个元素都使用相同的 v 计算理论价

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 v 中对应的值计算理论价

        t (float / pandas.Series): 年化到期时间，例如：还有 100 天到期，则年化到期时间为 100/360

            * float: 对于 series 中每个元素都使用相同的 t 计算理论价

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 t 中对应的值计算理论价

        option_class (str / pandas.Series): 期权方向，必须是两者其一，否则返回的序列值全部为 nan

            * str: "CALL" 或者 "PUT"

            * pandas.Series: 其元素个数应该和 series 元素个数相同，每个元素的值为 "CALL" 或者 "PUT"

        d1 (None | pandas.Series): [可选] 序列对应的 BS 公式中 b1 值

    Returns:
        pandas.Series: 该序列的 rho 值


    Example::

        import pandas as pd
        from tqsdk import TqApi, TqAuth, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        quote = api.get_quote("SHFE.cu2006")
        ks = api.get_kline_serial("SHFE.cu2006", 24 * 60 * 60, 10)
        v = tafunc.get_his_volatility(ks, quote)  # 历史波动率

        option = api.get_quote("SHFE.cu2006C45000")
        klines = api.get_kline_serial(["SHFE.cu2006C45000", "SHFE.cu2006"], 24 * 60 * 60, 10)
        t = tafunc.get_t(klines, option.expire_datetime)
        impv = tafunc.get_impv(klines["close1"], klines["close"], 45000, 0.025, v, t, "CALL")
        rho = tafunc.get_rho(klines["close1"], 45000, 0.025, v, t, "CALL")
        print("rho", list(rho))
        api.close()

    """
    o = _get_options_class(series, option_class=option_class)
    if d1 is None:
        d1 = _get_d1(series, k, r, v, t)
    d2 = pd.Series(np.where(np.isnan(d1), np.nan, d1 - v * np.sqrt(t)))
    return pd.Series(np.where(np.isnan(d1), np.nan, o * k * t * np.exp(-r * t) * _get_cdf(o * d2)))


def get_impv(series, series_option, k, r, init_v, t, option_class):
    """
    计算期权隐含波动率

    Args:
        series (pandas.Series): 标的价格序列

        series_option (pandas.Series): 期权价格序列，与 series 长度应该相同

        k (float): 期权行权价

        r (float): 无风险利率

        init_v (float / pandas.Series): 初始波动率，迭代初始值

            * float: 对于 series 中每个元素都使用相同的 init_v 计算隐含波动率

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 init_v 中对应的值计算隐含波动率

        t (float / pandas.Series): 年化到期时间，例如：还有 100 天到期，则年化到期时间为 100/360

            * float: 对于 series 中每个元素都使用相同的 t 计算理论价

            * pandas.Series: 其元素个数应该和 series 元素个数相同，对于 series 中每个元素都使用 t 中对应的值计算理论价

        option_class (str / pandas.Series): 期权方向，必须是两者其一，否则返回的序列值全部为 nan

            * str: "CALL" 或者 "PUT"

            * pandas.Series: 其元素个数应该和 series 元素个数相同，每个元素的值为 "CALL" 或者 "PUT"

    Returns:
        pandas.Series: 该序列的隐含波动率


    Example::

        import pandas as pd
        from tqsdk import TqApi, TqAuth, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        quote = api.get_quote("SHFE.cu2006")
        ks = api.get_kline_serial("SHFE.cu2006", 24 * 60 * 60, 10)
        v = tafunc.get_his_volatility(ks, quote)  # 历史波动率

        option = api.get_quote("SHFE.cu2006C45000")
        klines = api.get_kline_serial(["SHFE.cu2006C45000", "SHFE.cu2006"], 24 * 60 * 60, 10)
        t = tafunc.get_t(klines, option.expire_datetime)
        impv = tafunc.get_impv(klines["close1"], klines["close"], 45000, 0.025, v, t, "CALL")
        print("impv", list((impv * 100).round(2)))
        api.close()
    """
    o = _get_options_class(series, option_class=option_class)
    lower_limit = o * (series - k * np.exp(-r * t))
    x = pd.Series(np.where((series_option < lower_limit) | (t <= 0), np.nan, init_v))
    y = pd.Series(np.where(np.isnan(x), np.nan, get_bs_price(series, k, r, x, t, option_class)))
    vega = get_vega(series, k, r, x, t)
    diff_x = pd.Series(np.where(np.isnan(vega) | (vega < 1e-8), np.nan, (series_option - y) / vega))
    while not pd.DataFrame.all((np.abs(series_option - y) < 1e-8) | np.isnan(diff_x)):
        x = pd.Series(np.where(np.isnan(x) | np.isnan(diff_x), x,
                               np.where(x + diff_x < 0, x / 2,
                                        np.where(diff_x > x / 2, x * 1.5, x + diff_x))))
        y = pd.Series(np.where(np.isnan(x), np.nan, get_bs_price(series, k, r, x, t, option_class)))
        vega = get_vega(series, k, r, x, t)
        diff_x = pd.Series(np.where(np.isnan(vega) | (vega < 1e-8), np.nan, (series_option - y) / vega))
    return x


def get_ticks_info(df):
    """
    计算 ticks 开平方向

    Args:
        df (pandas.DataFrame): Dataframe 格式的 ticks 序列

    Returns:
        pandas.Series: 返回序列的开平方向序列

    Example::

        from tqsdk import TqApi, TqAuth, tafunc

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))
        ticks = api.get_tick_serial('SHFE.cu2006', 100)
        ticksinfo = tafunc.get_ticks_info(ticks)
        for i, v in ticksinfo.items():
            print(f"{tafunc.time_to_str(ticks['datetime'][i])[5:21]}  {ticks['last_price'][i]}  {v}")
        api.close()

        # 预计的输出是这样的:
        04-27 10:54:11.5   42640.0   多换
        04-27 10:54:12.0   42640.0   多换
        04-27 10:54:16.5   42640.0   多换
        ......
        04-27 10:55:10.0   42660.0   双平
        04-27 10:55:10.5   42660.0   双平
        04-27 10:55:14.0   42670.0   双平
    """
    if "open_interest" not in df.keys():  # df 不是 ticks 是 klines
        raise Exception(f"get_ticks_info 参数必须是 ticks，由 api.get_tick_serial 返回的对象。")
    df_pre = df.copy().shift(1)
    df_pre["price_diff"] = df["last_price"] - df_pre["last_price"]
    df_pre["oi_diff"] = df["open_interest"] - df_pre["open_interest"]
    df_pre["vol_diff"] = df["volume"] - df_pre["volume"]
    df_pre["pc"] = np.where(df["last_price"] <= df_pre["bid_price1"], -1,
                            np.where(df["last_price"] >= df_pre["ask_price1"], 1, np.sign(df_pre["price_diff"])))
    pc_g = df_pre["pc"] > 0
    df_pre["info"] = pd.Series(np.where(df_pre["oi_diff"] > 0, np.where(pc_g, "多开", "空开"),
                                        np.where(df_pre["oi_diff"] < 0, np.where(pc_g, "空平", "多平"),
                                                 np.where(df_pre["oi_diff"] == 0, np.where(pc_g, "多换", "空换"), ""))))
    df_pre.loc[df_pre["pc"] == 0, "info"] = "换手"
    df_pre.loc[(df_pre["oi_diff"] < 0) & (df_pre["oi_diff"] + df_pre["vol_diff"] == 0), "info"] = "双平"
    df_pre.loc[(df_pre["oi_diff"] > 0) & (df_pre["oi_diff"] == df_pre["vol_diff"]), "info"] = "双开"
    df_pre.loc[df_pre["vol_diff"] == 0, "info"] = ""
    return df_pre["info"]


def get_dividend_df(stock_dividend_ratio, cash_dividend_ratio):
    """
    计算复权系数矩阵

    Args:
        stock_dividend_ratio (list): 除权表（可以由 quote.stock_dividend_ratio 取得）

        cash_dividend_ratio (list): 除息表（可以由 quote.cash_dividend_ratio 取得）

    Returns:
        pandas.Dataframe: 复权系数矩阵， Dataframe 对象有 ["datetime", "stock_dividend", "cash_dividend"] 三列。
    """
    # 除权矩阵
    stock_dividend_df = pd.DataFrame({
        "datetime": [_datetime_to_timestamp_nano(datetime.datetime.strptime(s.split(",")[0], "%Y%m%d")) for s in
                     stock_dividend_ratio],
        "stock_dividend": np.array([float(s.split(",")[1]) for s in stock_dividend_ratio])
    })
    # 除息矩阵
    cash_dividend_df = pd.DataFrame({
        "datetime": [_datetime_to_timestamp_nano(datetime.datetime.strptime(s.split(",")[0], "%Y%m%d")) for s in
                     cash_dividend_ratio],
        "cash_dividend": [float(s.split(",")[1]) for s in cash_dividend_ratio]
    })
    # 除权除息矩阵
    dividend_df = pd.merge(stock_dividend_df, cash_dividend_df, on=['datetime'], how="outer", sort=True)
    dividend_df.fillna(0.0, inplace=True)
    return dividend_df


def get_dividend_factor(dividend_df, last_item, item):
    """
    返回 item 项对应的复权因子。

    Args:
        dividend_df (pandas.Dataframe): 除权除息矩阵表

        last_item (dict): 前一个 tickItem / klineItem

        item (dict): 当前 tickItem / klineItem

    Returns:
        float: 复权因子

    """
    last_dt = last_item['datetime']
    dt = item['datetime']
    if last_dt and dt:
        gt = dividend_df['datetime'].gt(last_dt)
        if gt.any():
            dividend_first = dividend_df[gt].iloc[0]
            if dt >= dividend_first['datetime']:
                c = last_item['close'] if last_item['close'] else last_item['last_price']
                return (1 - dividend_first['cash_dividend'] / c) / (1 + dividend_first['stock_dividend'])
    return 1


def _tq_pstdev(data: Series, mu: float):
    """
    计算标准差
    标准库提供的方法 statistics.pstdev 在 py3.6,py3.7 版本下参数 mean 不能设定为指定值，所以这里另外计算。
    """
    n = data.shape[0]
    assert n >= 1
    return math.sqrt(sum((data - mu)**2) / n)


def get_sharp(series, trading_days_of_year=250, r=0.025):
    """
    年化夏普率

    Args:
        series (pandas.Series): 每日收益率序列

        trading_days_of_year (int): 年化交易日数量

        r (float): 无风险利率

    Returns:
        float: 年化夏普率
    """
    rf = _get_daily_risk_free(trading_days_of_year, r)
    mean = series.mean()
    stddev = _tq_pstdev(series, mu=mean)
    return trading_days_of_year ** (1 / 2) * (mean - rf) / stddev if stddev else float("inf")


def get_sortino(series, trading_days_of_year=250, r=0.025):
    """
    年化索提诺比率

    Args:
        series (pandas.Series): 每日收益率序列

        trading_days_of_year (int): 年化交易日数量

        r (float): 无风险利率

    Returns:
        float: 年化索提诺比率
    """
    rf = _get_daily_risk_free(trading_days_of_year, r)
    mean = series.mean()
    left_daily_yield = series.loc[series < rf]
    stddev = _tq_pstdev(left_daily_yield, mu=rf) if left_daily_yield.shape[0] > 0 else 0
    return (trading_days_of_year * left_daily_yield.shape[0] / series.shape[0]) ** (1 / 2) * (mean - rf) / stddev if stddev else float("inf")


def get_calmar(series, max_drawdown, trading_days_of_year=250, r=0.025):
    """
    年化卡玛比率

    Args:
        series (pandas.Series): 每日收益率序列

        max_drawdown (float): 最大回撤

        trading_days_of_year (int): 年化交易日数量

        r (float): 无风险利率

    Returns:
        float: 年化夏普率
    """
    rf = _get_daily_risk_free(trading_days_of_year, r)
    if max_drawdown and max_drawdown == max_drawdown:
        mean = series.mean()
        return trading_days_of_year ** (1 / 2) * (mean - rf) / max_drawdown
    return float("inf")


def _get_daily_risk_free(trading_days_of_year, r):
    """日化无风险利率"""
    return pow(r + 1, 1 / trading_days_of_year) - 1


def _cum_counts(s: Series):
    """
    统计连续为1的个数, 用于计算最大连续盈利/亏损天数
    input:  [0, 0, 1, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 0]
    output: [0, 0, 1, 0, 1, 0, 1, 0, 0, 1, 2, 0, 1, 0, 0, 1, 2, 3, 4, 0, 1, 2, 3, 4, 5, 0, 1, 2, 3, 0, 0]
    """
    return s * (s.groupby((s != s.shift()).cumsum()).cumcount() + 1)




#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = "mayanqiong"

from tqsdk.tradeable.sim.trade_base import SimTradeBase
from tqsdk.tradeable.sim.utils import _get_stock_fee, _get_order_price, _get_dividend_ratio


class SimTradeStock(SimTradeBase):
    """
    天勤模拟交易账户，期货及商品期权
    """

    def _generate_account(self, init_balance):
        return {
            "user_id": self._account_id,  # 客户号, 与 order / trade 对象中的 user_id 值保持一致
            "currency": "CNY",
            "market_value_his": 0.0,  # 期初市值
            "asset_his": init_balance,  # 期初资产
            "cost_his": 0.0,  # 期初买入成本
            "deposit": 0.0,
            "withdraw": 0.0,
            "dividend_balance_today": 0.0,  # 当日分红金额

            "available_his": init_balance,

            "market_value": 0.0,  # 当前市值
            "asset": init_balance,  # 当前资产 = 当前市值 + 可用余额 + 冻结
            "available": init_balance,  # 可用余额 = 期初余额 + 当日分红金额 - 买入费用 - 卖出费用 + 当日入金 - 当日出金 - 当日买入占用资金 + 当日卖出释放资金 - 委托冻结金额 - 委托冻结费用
            "drawable": init_balance,  # 可取余额 = 可用余额 - 当日卖出释放资金
            "buy_frozen_balance": 0.0,  # 当前交易冻结金额（不含费用）= sum(order.volume_orign * order.limit_price)
            "buy_frozen_fee": 0.0,  # 当前交易冻结费用 = sum(order.frozen_fee)
            "buy_balance_today": 0.0,  # 当日买入占用资金（不含费用）
            "buy_fee_today": 0.0,  # 当日买入累计费用
            "sell_balance_today": 0.0,  # 当日卖出释放资金
            "sell_fee_today": 0.0,  # 当日卖出累计费用
            "cost": 0.0,  # 当前买入成本 = SUM（买入成本）
            "hold_profit": 0.0,  # 当日持仓盈亏 = 当前市值 - 当前买入成本
            "float_profit_today": 0.0,  # 当日浮动盈亏 = SUM(持仓当日浮动盈亏)
            "real_profit_today": 0.0,  # 当日实现盈亏 = SUM(持仓当日实现盈亏)
            "profit_today": 0.0,  # 当日盈亏 = 当日浮动盈亏 + 当日实现盈亏
            "profit_rate_today": 0.0  # 当日盈亏比 = 当日盈亏 / (当前买入成本 if 当前买入成本 > 0 else 期初资产)
        }

    def _generate_position(self, symbol, quote, underlying_quote) -> dict:
        return {
            "user_id": self._account_id,
            "exchange_id": symbol.split(".", maxsplit=1)[0],
            "instrument_id": symbol.split(".", maxsplit=1)[1],
            "create_date": "",  # 建仓日期
            "volume_his": 0,  # 昨持仓数量
            "cost_his": 0.0,  # 期初买入成本
            "market_value_his": 0.0,  # 期初市值
            "real_profit_his": 0.0,  # 期初实现盈亏
            "shared_volume_today": 0,  # 今送股数量
            "devidend_balance_today": 0.0,  # 今分红金额

            "buy_volume_his": 0,  # 期初累计买入持仓
            "buy_balance_his": 0.0,  # 期初累计买入金额
            "buy_fee_his": 0.0,  # 期初累计买入费用
            "sell_volume_his": 0,  # 期初累计卖出持仓
            "sell_balance_his": 0.0,  # 期初累计卖出金额
            "sell_fee_his": 0.0,  # 期初累计卖出费用

            "buy_volume_today": 0,  # 当日累计买入持仓
            "buy_balance_today": 0.0,  # 当日累计买入金额 (不包括费用)
            "buy_fee_today": 0.0,  # 当日累计买入费用
            "sell_volume_today": 0,  # 当日累计卖出持仓
            "sell_balance_today": 0.0,  # 当日累计卖出金额 (不包括费用)
            "sell_fee_today": 0.0,  # 当日累计卖出费用

            "last_price": quote["last_price"],
            "sell_volume_frozen": 0,  # 今日卖出冻结手数
            "sell_float_profit_today": 0.0,  # 昨仓浮动盈亏 = (昨持仓数量 - 今卖数量) * (最新价 - 昨收盘价)
            "buy_float_profit_today": 0.0,  # 今仓浮动盈亏 = (今持仓数量 - (昨持仓数量 - 今卖数量)) * (最新价 - 买入均价)
            # 买入均价 = (buy_balance_today + buy_fee_today) / buy_volume_today

            "cost": 0.0,  # 当前成本 = 期初成本 + 今买金额 + 今买费用 - 今卖数量 × (期初买入成本 / 期初持仓数量)
            "volume": 0,  # 今持仓数量 = 昨持仓数量 + 今买数量 - 今卖数量 + 送股数量
            "market_value": 0.0,  # 当前市值 = 持仓数量 × 行情最新价
            "float_profit_today": 0.0,  # 当日浮动盈亏 = sell_float_profit_today + buy_float_profit_today
            "real_profit_today": 0.0,  # 当日实现盈亏 = 今卖数量 * (最新价 - 昨收盘价) - 今卖费用 + 今派息金额
            "profit_today": 0.0,  # 当日盈亏 = 当日浮动盈亏 + 当日实现盈亏
            "profit_rate_today": 0.0,  # 当日收益率 = 当日盈亏 / ( 当前成本 if 当前成本 > 0 else 期初市值)
            "hold_profit": 0.0,  # 当日持仓盈亏 = 当前市值 – 当前买入成本
            "real_profit_total": 0.0,  # 累计实现盈亏 += 当日实现盈亏（成本）
            "profit_total": 0.0,  # 总盈亏 = 累计实现盈亏 + 持仓盈亏
            "profit_rate_total": 0.0,  # 累计收益率 = 总盈亏 / (当前成本 if 当前成本 > 0 else 期初成本)
        }

    def _generate_order(self, pack: dict) -> dict:
        """order 对象预处理"""
        order = pack.copy()
        order["exchange_order_id"] = order["order_id"]
        order["volume_orign"] = order["volume"]
        order["volume_left"] = order["volume"]
        order["frozen_balance"] = 0.0
        order["frozen_fee"] = 0.0
        order["last_msg"] = "报单成功"
        order["status"] = "ALIVE"
        order["insert_date_time"] = self._get_trade_timestamp()
        del order["aid"]
        del order["volume"]
        self._append_to_diffs(["orders", order["order_id"]], order)
        return order

    def _generate_trade(self, order, quote, price) -> dict:
        fee = _get_stock_fee(order["direction"], order["volume_left"], price)
        return {
            "user_id": order["user_id"],
            "order_id": order["order_id"],
            "trade_id": order["order_id"] + "|" + str(order["volume_left"]),
            "exchange_trade_id": order["order_id"] + "|" + str(order["volume_left"]),
            "exchange_id": order["exchange_id"],
            "instrument_id": order["instrument_id"],
            "direction": order["direction"],  # 下单方向, BUY=买, SELL=卖，SHARED=送股，DEVIDEND=分红 (送股|分红没有计算费用)
            "price": price,
            "volume": order["volume_left"],
            "trade_date_time": self._get_trade_timestamp(),  # todo: 可能导致测试结果不确定
            "fee": fee
        }

    def _on_settle(self):
        for symbol in self._orders:
            for order in self._orders[symbol].values():
                order["frozen_balance"] = 0.0
                order["frozen_fee"] = 0.0
                order["last_msg"] = "交易日结束，自动撤销当日有效的委托单（GFD）"
                order["status"] = "FINISHED"
                self._append_to_diffs(["orders", order["order_id"]], order)

        dividend_balance_today = 0.0  # 今日分红总的分红数据
        for position in self._positions.values():
            symbol = f"{position['exchange_id']}.{position['instrument_id']}"
            quote, _ = self._get_quotes_by_symbol(symbol)
            stock_dividend, cash_dividend = _get_dividend_ratio(quote)
            # position 原始字段
            position["volume_his"] = position["volume"]  # 期初持仓数量
            position["cost_his"] = position["cost"]  # 期初买入成本
            position["market_value_his"] = position["market_value"]  # 期初市值
            position["real_profit_his"] = position["real_profit_today"]  # 期初实现盈亏

            # 处理分红送股
            position["shared_volume_today"] = stock_dividend * position["volume"]  # 今送股数量
            position["devidend_balance_today"] = cash_dividend * position["volume"]  # 今分红金额
            if position["shared_volume_today"] > 0.0 or position["devidend_balance_today"] > 0.0:
                position["volume"] += position["shared_volume_today"]
                position["market_value"] -= position["devidend_balance_today"]  # 分红后的市值
                position["last_price"] = position["market_value"] / position["volume"]  # 分红送股后的最新价, todo: 可能会于第二天收到的第一笔行情有误差？
                dividend_balance_today += position["devidend_balance_today"]  # 记录累积分红金额，account 需要

            position["buy_volume_his"] = position["buy_volume_today"]
            position["buy_balance_his"] = position["buy_balance_today"]
            position["buy_fee_his"] = position["buy_fee_today"]
            position["sell_volume_his"] = position["sell_volume_today"]
            position["sell_balance_his"] = position["sell_balance_today"]
            position["sell_fee_his"] = position["sell_fee_today"]
            position["buy_volume_today"] = 0
            position["buy_balance_today"] = 0.0
            position["buy_fee_today"] = 0.0
            position["sell_volume_today"] = 0
            position["sell_balance_today"] = 0.0
            position["sell_fee_today"] = 0.0

            position["sell_volume_frozen"] = 0
            position["buy_avg_price"] = 0.0
            position["sell_float_profit_today"] = 0.0
            position["buy_float_profit_today"] = 0.0

            position["float_profit_today"] = 0.0  # 当日浮动盈亏 = position["sell_float_profit_today"] + position["buy_float_profit_today"]
            position["real_profit_today"] = 0.0  # 当日实现盈亏 = 今卖数量 * (最新价 - 昨收盘价) - 今卖费用 + 今派息金额
            position["profit_today"] = 0.0  # 当日盈亏 = 当日浮动盈亏 + 当日实现盈亏
            position["profit_rate_today"] = 0.0 # 当日收益率 = 当日盈亏 / ( 当前成本 if 当前成本 > 0 else 期初市值)
            position["hold_profit"] = 0.0  # 当日持仓盈亏 = 当前市值 – 当前买入成本
            self._append_to_diffs(["positions", symbol], position)

        # account 原始字段
        self._account["dividend_balance_today"] = dividend_balance_today
        self._account["market_value_his"] = self._account["market_value"]
        self._account["asset_his"] = self._account["asset"]
        self._account["cost_his"] = self._account["cost"]
        self._account["available_his"] = self._account["available"] + self._account["buy_frozen_balance"] + self._account["buy_frozen_fee"]
        self._account["buy_frozen_balance"] = 0.0
        self._account["buy_frozen_fee"] = 0.0
        self._account["buy_balance_today"] = 0.0
        self._account["buy_fee_today"] = 0.0
        self._account["sell_balance_today"] = 0.0
        self._account["sell_fee_today"] = 0.0
        self._account["asset"] += self._account["dividend_balance_today"]
        self._account["market_value"] -= self._account["dividend_balance_today"]
        # account 计算字段
        self._account["available"] = self._account["asset"] - self._account["market_value"]  # 当前可用余额 = 当前资产 - 当前市值
        self._account["drawable"] = self._account["available"]
        self._account["hold_profit"] = 0.0  # 当日持仓盈亏 = 当前市值 - 当前买入成本
        self._account["float_profit_today"] = 0.0  # 当日浮动盈亏 = SUM(持仓当日浮动盈亏)
        self._account["real_profit_today"] = 0.0  # 当日实现盈亏 = SUM(持仓当日实现盈亏)
        self._account["profit_today"] = 0.0  # 当日盈亏 = 当日浮动盈亏 + 当日实现盈亏
        self._account["profit_rate_today"] = 0.0  # 当日盈亏比 = 当日盈亏 / (当前买入成本 if 当前买入成本 > 0 else 期初资产)
        # 根据公式 账户权益 不需要计算 self._account["balance"] = static_balance + market_value
        self._append_to_diffs(["accounts", "CNY"], self._account)

    def _check_insert_order(self, order, symbol, position, quote, underlying_quote=None):
        # 无法计入 orderbook
        if quote["ins_class"] != "STOCK":
            order["last_msg"] = "不支持的合约类型，TqSimStock 只支持股票模拟交易"
            order["status"] = "FINISHED"

        if order["status"] == "ALIVE" and not self._is_in_trading_time(quote):
            order["last_msg"] = "下单失败, 不在可交易时间段内"
            order["status"] = "FINISHED"

        if order["status"] == "ALIVE" and order["direction"] == "BUY":
            price = _get_order_price(quote, order)
            order["frozen_balance"] = price * order["volume_orign"]
            order["frozen_fee"] = _get_stock_fee(order["direction"], order["volume_orign"], price)
            if order["frozen_balance"] + order["frozen_fee"] > self._account["available"]:
                order["frozen_balance"] = 0.0
                order["frozen_fee"] = 0.0
                order["last_msg"] = "开仓资金不足"
                order["status"] = "FINISHED"

        if order["status"] == "ALIVE" and order["direction"] == "SELL":
            if position["volume_his"] + position["shared_volume_today"] - position["sell_volume_today"] - position["sell_volume_frozen"] < order["volume_orign"]:
                order["last_msg"] = "平仓手数不足"
                order["status"] = "FINISHED"

        if order["status"] == "FINISHED":
            self._append_to_diffs(["orders", order["order_id"]], order)

    def _on_insert_order(self, order, symbol, position, quote, underlying_quote=None):
        """记录在 orderbook"""
        if order["direction"] == "BUY":
            self._adjust_account_by_order(buy_frozen_balance=order["frozen_balance"], buy_frozen_fee=order["frozen_fee"])
            self._append_to_diffs(["accounts", "CNY"], self._account)
        else:
            position["sell_volume_frozen"] += order["volume_orign"]
            self._append_to_diffs(["positions", symbol], position)

    def _on_order_failed(self, symbol, order):
        origin_frozen_balance = order["frozen_balance"]
        origin_frozen_fee = order["frozen_fee"]
        order["frozen_balance"] = 0.0
        order["frozen_fee"] = 0.0
        self._append_to_diffs(["orders", order["order_id"]], order)
        # 调整账户和持仓
        if order["direction"] == "BUY":
            self._adjust_account_by_order(buy_frozen_balance=-origin_frozen_balance, buy_frozen_fee=-origin_frozen_fee)
            self._append_to_diffs(["accounts", "CNY"], self._account)
        else:
            position = self._positions[symbol]
            position["sell_volume_frozen"] -= order["volume_orign"]
            self._append_to_diffs(["positions", symbol], position)

    def _on_order_traded(self, order, trade, symbol, position, quote, underlying_quote):
        origin_frozen_balance = order["frozen_balance"]
        origin_frozen_fee = order["frozen_fee"]
        order["frozen_balance"] = 0.0
        order["frozen_fee"] = 0.0
        order["volume_left"] = 0
        self._append_to_diffs(["trades", trade["trade_id"]], trade)
        self._append_to_diffs(["orders", order["order_id"]], order)

        # 调整账户和持仓
        if order["direction"] == "BUY":
            if position["volume"] == 0:
                position["create_date"] = quote['datetime'][:10]
            self._adjust_account_by_order(buy_frozen_balance=-origin_frozen_balance, buy_frozen_fee=-origin_frozen_fee)
            # 修改 position 原始字段
            buy_balance = trade["volume"] * trade["price"]
            position["buy_volume_today"] += trade["volume"]
            position["buy_balance_today"] += buy_balance
            position["buy_fee_today"] += trade["fee"]
            # 修改 account 原始字段
            self._adjust_account_by_trade(buy_fee=trade["fee"], buy_balance=buy_balance)
            self._adjust_position_account(position, pre_last_price=trade["price"], last_price=position["last_price"],
                                          buy_volume=trade["volume"], buy_balance=buy_balance, buy_fee=trade["fee"])
        else:
            position["sell_volume_frozen"] -= order["volume_orign"]
            # 修改 position 原始字段
            sell_balance = trade["volume"] * trade["price"]
            position["sell_volume_today"] += trade["volume"]
            position["sell_balance_today"] += sell_balance
            position["sell_fee_today"] += trade["fee"]
            self._adjust_account_by_trade(sell_fee=trade["fee"], sell_balance=sell_balance)
            self._adjust_position_account(position, last_price=quote["last_price"], sell_volume=trade["volume"],
                                          sell_balance=sell_balance, sell_fee=trade["fee"])

        self._append_to_diffs(["positions", symbol], position)
        self._append_to_diffs(["accounts", "CNY"], self._account)

    def _on_update_quotes(self, symbol, position, quote, underlying_quote):
        # 调整持仓保证金和盈亏
        if position["volume"] > 0:
            if position["last_price"] != quote["last_price"]:
                self._adjust_position_account(position, pre_last_price=position["last_price"], last_price=quote["last_price"])
                position["last_price"] = quote["last_price"]
        # 修改辅助变量
        position["last_price"] = quote["last_price"]
        self._append_to_diffs(["positions", symbol], position)  # 一定要返回 position，下游会用到 future_margin 字段判断修改保证金是否成功
        self._append_to_diffs(["accounts", "CNY"], self._account)

    def _adjust_position_account(self, position, pre_last_price=float("nan"), last_price=float("nan"), buy_volume=0, buy_balance=0, buy_fee=0, sell_volume=0, sell_balance=0, sell_fee=0):
        """
        价格变化，使得 position 中的以下计算字段需要修改，这个函数计算出需要修改的差值部分，计算出差值部分修改 position、account
        有两种情况下调用
        1. 委托单 FINISHED，且全部成交，分为4种：buy_open, buy_close, sell_open, sell_close
        2. 行情跳动
        """
        assert [buy_volume, sell_volume].count(0) >= 1  # 只有一个大于0, 或者都是0，表示价格变化导致的字段修改
        if buy_volume > 0:
            position["volume"] += buy_volume
            cost = buy_balance + buy_fee
            market_value = buy_volume * position["last_price"]
            position["buy_avg_price"] = (position["buy_balance_today"] + position["buy_fee_today"]) / position["buy_volume_today"]
            buy_float_profit_today = (position["volume"] - (position["volume_his"] - position["sell_volume_today"])) * (last_price - position["buy_avg_price"])  # 今仓浮动盈亏 = (今持仓数量 - (昨持仓数量 - 今卖数量)) * (最新价 - 买入均价)
            self._adjust_position(position, cost=cost, market_value=market_value,
                                  sell_float_profit_today=0,
                                  buy_float_profit_today=buy_float_profit_today, real_profit_today=0)
            self._adjust_account_by_position(market_value=market_value, cost=cost,
                                             float_profit_today=buy_float_profit_today,
                                             real_profit_today=0)
        elif sell_volume > 0:
            position["volume"] -= sell_volume
            cost = -sell_volume * (position["cost_his"] / position["volume_his"])
            market_value = -sell_volume * position["last_price"]
            real_profit_today = (sell_volume / position["volume_his"]) * position["sell_float_profit_today"]
            sell_float_profit_today = position["sell_float_profit_today"] - real_profit_today
            self._adjust_position(position, cost=cost, market_value=market_value,
                                  sell_float_profit_today=sell_float_profit_today,
                                  buy_float_profit_today=0, real_profit_today=real_profit_today)
            self._adjust_account_by_position(market_value=market_value, cost=cost,
                                             float_profit_today=sell_float_profit_today,
                                             real_profit_today=real_profit_today)
        else:
            market_value = position["volume"] * last_price - position["market_value"]
            sell_float_profit_today = (position["volume_his"] - position["sell_volume_today"]) * (last_price - pre_last_price)  # 昨仓浮动盈亏 = (昨持仓数量 - 今卖数量) * (最新价 - 昨收盘价)
            buy_float_profit_today = (position["volume"] - (position["volume_his"] - position["sell_volume_today"])) * (last_price - position["buy_avg_price"])  # 今仓浮动盈亏 = (今持仓数量 - (昨持仓数量 - 今卖数量)) * (最新价 - 买入均价)
            self._adjust_position(position, cost=0, market_value=market_value, sell_float_profit_today=sell_float_profit_today, buy_float_profit_today=buy_float_profit_today, real_profit_today=0)
            self._adjust_account_by_position(market_value=market_value, cost=0, float_profit_today=sell_float_profit_today+buy_float_profit_today, real_profit_today=0)

    # -------- 对于 position 的计算字段修改分为两类：
    # 1. 针对手数相关的修改，在下单、成交时会修改
    # 2. 针对盈亏、保证金、市值的修改，由于参考合约最新价，在成交、行情跳动时会修改
    def _adjust_position(self, position, cost=0, market_value=0, sell_float_profit_today=0, buy_float_profit_today=0, real_profit_today=0):
        # 更新 position 计算字段，根据差值更新的字段
        position["sell_float_profit_today"] += sell_float_profit_today
        position["buy_float_profit_today"] += buy_float_profit_today

        position["cost"] += cost  # 当前成本 = 期初成本 + 今买金额 + 今买费用 - 今卖数量 × (期初买入成本 / 期初持仓数量)
        position["market_value"] += market_value  # 当前市值 = 持仓数量 × 行情最新价
        position["float_profit_today"] += sell_float_profit_today + buy_float_profit_today  # 当日浮动盈亏 = (昨持仓数量 - 今卖数量) * (最新价 - 昨收盘价) + (今持仓数量 - (昨持仓数量 - 今卖数量)) * (最新价 - 买入均价)
        position["real_profit_today"] += real_profit_today  # 当日实现盈亏 = 今卖数量 * (最新价 - 昨收盘价) - 今卖费用 + 今派息金额
        position["profit_today"] += sell_float_profit_today + buy_float_profit_today + real_profit_today
        position["hold_profit"] += (market_value - cost)
        position["real_profit_total"] += real_profit_today  # 累计实现盈亏 += 当日实现盈亏（成本）
        position["profit_total"] += real_profit_today + (market_value - cost)  # 总盈亏 = 累计实现盈亏 + 持仓盈亏
        # 当日收益率 = 当日盈亏 / ( 当前成本 if 当前成本 > 0 else 期初市值)
        if position["cost"] > 0:
            position["profit_rate_today"] = position["profit_today"] / position["cost"]
        else:
            position["profit_rate_today"] = position["profit_today"] / position["market_value_his"] if position["market_value_his"] > 0 else 0.0
        # 累计收益率 = 总盈亏 / (当前成本 if 当前成本 > 0 else 期初成本)
        if position["cost"] > 0:
            position["profit_rate_total"] = position["profit_total"] / position["cost"]
        else:
            position["profit_rate_total"] = position["profit_total"] / position["cost_his"] if position["cost_his"] > 0 else 0.0


    # -------- 对于 account 的修改分为以下三类
    def _adjust_account_by_trade(self, buy_fee=0, buy_balance=0, sell_fee=0, sell_balance=0):
        """由成交引起的 account 原始字段变化，account 需要更新的计算字段"""
        # account 原始字段
        self._account["buy_balance_today"] += buy_balance  # 当日买入占用资金（不含费用）
        self._account["buy_fee_today"] += buy_fee  # 当日买入累计费用
        self._account["sell_balance_today"] += sell_balance  # 当日卖出释放资金
        self._account["sell_fee_today"] += sell_fee  # 当日卖出累计费用
        # account 计算字段
        self._account["available"] += (sell_balance - buy_fee - sell_fee - buy_balance)
        self._account["asset"] += (sell_balance - buy_fee - sell_fee - buy_balance)
        self._account["drawable"] = max(self._account["available_his"] + min(0, self._account["sell_balance_today"] - self._account["buy_balance_today"] - self._account["buy_fee_today"] - self._account["buy_frozen_balance"] - self._account["buy_frozen_fee"]), 0)

    def _adjust_account_by_position(self, market_value=0, cost=0, float_profit_today=0, real_profit_today=0):
        """由 position 变化，account 需要更新的计算字段"""
        # account 计算字段，持仓字段求和的字段
        self._account["market_value"] += market_value
        self._account["cost"] += cost
        self._account["float_profit_today"] += float_profit_today
        self._account["real_profit_today"] += real_profit_today
        # account 计算字段
        self._account["asset"] += market_value  # 当前资产 = 当前市值 + 当前可用余额 + 委托冻结金额 + 委托冻结费用
        # 当前可取余额 = MAX( 期初余额 + MIN(0，当日卖出释放资金 - 当日买入占用资金 - 委托冻结金额) , 0)
        self._account["drawable"] = max(self._account["available_his"] + min(0, self._account["sell_balance_today"] - self._account["buy_balance_today"] - self._account["buy_fee_today"] - self._account["buy_frozen_balance"] - self._account["buy_frozen_fee"]), 0)
        self._account["hold_profit"] = self._account["market_value"] - self._account["cost"]  # 当日持仓盈亏 = 当前市值 - 当前买入成本
        self._account["profit_today"] = self._account["float_profit_today"] + self._account["real_profit_today"]  # 当日盈亏 = 当日浮动盈亏 + 当日实现盈亏
        # 当日盈亏比 = 当日盈亏 / (当前买入成本 if 当前买入成本 > 0 else 期初资产)
        if self._account["cost"] > 0:
            self._account["profit_rate_today"] = self._account["profit_today"] / self._account["cost"]
        else:
            self._account["profit_rate_today"] = self._account["profit_today"] / self._account["asset_his"] if self._account["asset_his"] > 0 else 0.0

    def _adjust_account_by_order(self, buy_frozen_balance=0, buy_frozen_fee=0):
        """由 order 变化，account 需要更新的计算字段"""
        self._account["buy_frozen_balance"] += buy_frozen_balance
        self._account["buy_frozen_fee"] += buy_frozen_fee
        self._account["available"] -= (buy_frozen_balance + buy_frozen_fee)
        self._account["drawable"] = max(self._account["available_his"] + min(0, self._account["sell_balance_today"] - self._account["buy_balance_today"] - self._account["buy_fee_today"] - self._account["buy_frozen_balance"] - self._account["buy_frozen_fee"]), 0)



#!usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'mayanqiong'

from tqsdk.datetime import _get_trading_day_from_timestamp, _get_trading_day_end_time
from tqsdk.datetime_state import TqDatetimeState
from tqsdk.exceptions import TqRiskRuleError


class TqRiskManager(list):

    def __init__(self):
        self._datetime_state = TqDatetimeState()
        self._trading_day_end = 0
        super(TqRiskManager, self).__init__()

    def _on_recv_data(self, diffs):
        for d in diffs:
            self._datetime_state.update_state(d)

        current = self._datetime_state.get_current_dt()
        if current > self._trading_day_end:
            # 切换交易日
            self._trading_day_end = _get_trading_day_end_time(_get_trading_day_from_timestamp(current))
            [r._on_settle() for r in self]

    def append(self, rule):
        if rule not in self:
            super(TqRiskManager, self).append(rule)

    def remove(self, rule):
        if rule in self:
            super(TqRiskManager, self).remove(rule)

    def _could_insert_order(self, pack):
        # 是否可以下单
        for r in self:
            is_valid, err_msg = r._could_insert_order(pack)
            if not is_valid:
                raise TqRiskRuleError(err_msg)
        return True

    def _on_insert_order(self, pack):
        # 需要更新风控对象内部统计值
        for r in self:
            r._on_insert_order(pack)



#!usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'mayanqiong'

import math
from abc import abstractmethod
from datetime import datetime
from typing import Callable

from tqsdk.datetime import _is_in_trading_time, _str_to_timestamp_nano
from tqsdk.diff import _simple_merge_diff
from tqsdk.tradeable.sim.utils import _get_price_range


class SimTradeBase(object):
    """
    本模块为 TqSim 交易部分的子模块的基类，纯同步计算，不涉及连接行情的动态信息，所以对于接口的调用有一些需要注意的要求

    不同的账户类型继承本模块，并实现具体账户类型的计算

    对外提供的接口:

    + init_snapshot: 返回初始的账户截面信息
    + insert_order -> (diffs, orders_events): 处理下单请求，调用 TqSimAccount.insert_order 之前应该调用过 update_quote，保证收到过合约的行情；期权还应该确保收到了标的的行情
    + cancel_order -> (diffs, orders_events)：处理撤单请求
    + update_quote -> (diffs, orders_events)：处理行情更新
    + settle -> (diffs, _orders_events, trade_log)：处理结算请求

    注意：

    + diffs (list) : 每个元素都是符合 diff 协议中 trade 交易部分的数据包，且返回的都是完整字段的对象，比如：order成交时，返回的是order完整对象而不是有变化的字段
    + orders_events  (list) : 按照顺序记录 order 的更新，返回给调用方
    + trade_log (dict) : 是结算前的账户日志信息

    diffs 由子类负责调用 _append_to_diffs
    orders_events 由父类统一处理在 order 状态变化时，在 list 中追加委托单实例

    """

    def __init__(self, account_key: str, account_id: str = "", init_balance: float = 10000000.0,
                 get_trade_timestamp: Callable = None, is_in_trading_time: Callable = None) -> None:
        self._account_key = account_key
        self._account_id = account_id
        self._quotes = {}  # 会记录所有的发来的行情
        # 初始化账户结构
        self._account = self._generate_account(init_balance)
        self._positions = {}  # {symbol: position, ...}
        self._orders = {}  # {symbol: {order_id: order}, ...}
        self._trades = []  # list 类型，与重构之前代码保持一致，list 可以保留 trade 生产的顺序信息
        self._diffs = []
        self._orders_events = []  # 按照顺序记录 order 的更新，返回给调用方
        self._max_datetime = ""  # 所有 quotes 的最大行情更新时间
        # 本模块在计算成交时间、判断是否在交易时间段内，默认使用所有 quotes 的最大行情更新时间当作当前时间，并且没有模拟到交易时的时间差
        # 若外部调用模块需要更精确时间，则由外部模块提供函数支持
        self._get_trade_timestamp = get_trade_timestamp if get_trade_timestamp else self._default_get_trade_timestamp
        self._is_in_trading_time = is_in_trading_time if is_in_trading_time else self._default_is_in_trading_time

    @abstractmethod
    def _generate_account(self, init_balance: float) -> dict:
        """返回 account 账户结构"""
        pass

    @abstractmethod
    def _generate_order(self, pack: dict) -> dict:
        """返回 order 委托单结构"""
        pass

    @abstractmethod
    def _generate_trade(self, order, quote, price) -> dict:
        """返回 trade 成交对象结构"""
        pass

    @abstractmethod
    def _generate_position(self, symbol, quote, underlying_quote) -> dict:
        """返回 position 对象结构"""
        pass

    @abstractmethod
    def _check_insert_order(self, order, symbol, position, quote, underlying_quote):
        """检查是否可以下单，在 order 原对象上修改属性"""
        pass

    @abstractmethod
    def _on_insert_order(self, order, symbol, position, quote, underlying_quote):
        """将 order 记入 order_book 时调用"""
        pass

    @abstractmethod
    def _on_order_failed(self, symbol, order):
        """处理 order 变为 FINISHED，且没有成交，撤单"""
        pass

    @abstractmethod
    def _on_order_traded(self, order, trade, symbol, position, quote, underlying_quote):
        """处理 order 变为 FINISHED，且全部成交"""
        pass

    @abstractmethod
    def _on_update_quotes(self, symbol, position, quote, underlying_quote):
        """更新合约行情后，更行对应的持仓及账户信息"""
        pass

    @abstractmethod
    def _on_settle(self):
        """结算时，应该调整 委托单、持仓、账户"""
        pass

    def init_snapshot(self):
        """返回初始账户截面信息"""
        return {
            "trade": {
                self._account_key: {
                    "accounts": {"CNY": self._account.copy()},
                    "positions": {},
                    "orders": {},
                    "trades": {}
                }
            }
        }

    def insert_order(self, symbol, pack):
        quote, underlying_quote = self._get_quotes_by_symbol(symbol)
        position = self._ensure_position(symbol, quote, underlying_quote)
        order = self._generate_order(pack)
        self._orders_events.append(order.copy())
        self._check_insert_order(order, symbol, position, quote, underlying_quote)
        if order["status"] == "FINISHED":
            self._orders_events.append(order.copy())
        else:
            orders = self._orders.setdefault(symbol, {})
            orders[order["order_id"]] = order  # order 存入全局
            self._on_insert_order(order, symbol, position, quote, underlying_quote)
            self._match_order(order, symbol, position, quote, underlying_quote)
        return self._return_results()

    def cancel_order(self, symbol, pack):
        order = self._orders.get(symbol, {}).get(pack["order_id"], {})
        if order.get("status") == "ALIVE":
            order["last_msg"] = "已撤单"
            order["status"] = "FINISHED"
            self._on_order_failed(symbol, order)
            self._orders_events.append(order)
            del self._orders[symbol][order["order_id"]]  # 删除 order
        return self._return_results()

    def update_quotes(self, symbol, pack):
        for q in pack.get("quotes", {}).values():
            self._max_datetime = max(q.get("datetime", ""), self._max_datetime)
        _simple_merge_diff(self._quotes, pack.get("quotes", {}))
        quote, underlying_quote = self._get_quotes_by_symbol(symbol)
        # 某些非交易时间段，ticks 回测是 quote 的最新价有可能是 nan，无效的行情直接跳过
        if math.isnan(quote["last_price"]):
            return [], []
        # 撮合委托单
        orders = self._orders.get(symbol, {})
        position = self._ensure_position(symbol, quote, underlying_quote)
        for order_id in list(orders.keys()):  # match_order 过程中可能会删除 orders 下对象
            self._match_order(orders[order_id], symbol, position, quote, underlying_quote)
        self._on_update_quotes(symbol, position, quote, underlying_quote)  # 调整持仓及账户信息
        return self._return_results()

    def _match_order(self, order, symbol, position, quote, underlying_quote=None):
        assert order["status"] == "ALIVE"
        status, last_msg, price = SimTradeBase.match_order(order, quote)
        if status == "FINISHED":
            order["last_msg"] = last_msg
            order["status"] = status
            if last_msg == "全部成交":
                trade = self._generate_trade(order, quote, price)
                self._trades.append(trade)
                self._on_order_traded(order, trade, symbol, position, quote, underlying_quote)
            else:
                self._on_order_failed(symbol, order)
            # 成交后记录 orders_event, 删除 order
            self._orders_events.append(order)
            del self._orders[symbol][order["order_id"]]

    def settle(self):
        # 生成 trade_log 时, 持仓信息中没有 pos_long、pos_short、pos，此处加补丁计算出来
        # 这几个计算值是在 data_extension 中计算的，只有在 api 里才能获取到，未来可以把 settle 的功能放在 api 中
        positions = {}
        for k, v in self._positions.items():
            positions[k] = v.copy()
            if v.get('pos_long_his') is not None and v.get('pos_long_today') is not None:
                positions[k]["pos_long"] = v['pos_long_his'] + v['pos_long_today']
            if v.get('pos_short_his') is not None and v.get('pos_short_today') is not None:
                positions[k]["pos_short"] = v['pos_short_his'] + v['pos_short_today']
            if v.get('pos_long') is not None and v.get('pos_short') is not None:
                positions[k]["pos"] = positions[k]["pos_long"] - positions[k]["pos_short"]
        trade_log = {
            "trades": self._trades,
            "account": self._account.copy(),
            "positions": positions
        }
        # 为下一交易日调整账户
        self._trades = []
        self._on_settle()
        for symbol in self._orders:
            for order in self._orders[symbol].values():
                self._orders_events.append(order)
            self._orders[symbol] = {}
        diffs, orders_events = self._return_results()
        return diffs, orders_events, trade_log

    def _ensure_position(self, symbol, quote, underlying_quote):
        position = self._positions.get(symbol, None)
        if position is None:
            position = self._generate_position(symbol, quote, underlying_quote)
            self._positions[symbol] = position
        return position

    def _get_quotes_by_symbol(self, symbol):
        """返回指定合约及标的合约，在本模块执行过程中，应该保证一定有合约行情"""
        quote = self._quotes.get(symbol)
        assert quote and quote.get("datetime"), "未收到指定合约行情"
        underlying_quote = None
        if quote["ins_class"].endswith("OPTION"):
            underlying_quote = self._quotes.get(quote["underlying_symbol"])
            assert underlying_quote and underlying_quote.get("datetime"), "未收到指定合约的标的行情"
        return quote, underlying_quote

    def _append_to_diffs(self, path, obj):
        target = {}
        diff = {'trade': {self._account_key: target}}
        while len(path) > 0:
            k = path.pop(0)
            target[k] = obj.copy() if len(path) == 0 else {}
            target = target[k]
        self._diffs.append(diff)

    def _return_results(self):
        """
        返回两项内容：diffs: list, orders_events: list
        diffs 是截面的变更
        orders_events 是委托单变化
        """
        diffs, self._diffs = self._diffs, []
        orders_events, self._orders_events = self._orders_events, []
        return diffs, orders_events

    def _default_get_trade_timestamp(self):
        """获取交易时间的默认方法，为当前所有 quote 的最大行情时间"""
        return _str_to_timestamp_nano(self._max_datetime)

    def _default_is_in_trading_time(self, quote):
        """判断是否在交易时间段"""
        return _is_in_trading_time(quote, self._max_datetime, float("nan"))

    @staticmethod
    def match_order(order, quote) -> (str, str, float):
        """
        撮合交易规则：
        * 市价单使用对手盘价格成交, 如果没有对手盘(涨跌停)则自动撤单
        * 限价单要求报单价格达到或超过对手盘价格才能成交, 成交价为报单价格, 如果没有对手盘(涨跌停)则无法成交
        * 模拟交易只有全部成交
        returns: status, last_msg, price
        """
        status, last_msg = "ALIVE", ""
        ask_price, bid_price = _get_price_range(quote)
        # order 预期成交价格
        if order["price_type"] in ["ANY", "BEST", "FIVELEVEL"]:
            price = ask_price if order["direction"] == "BUY" else bid_price
        else:
            price = order["limit_price"]
        if order["price_type"] == "ANY" and math.isnan(price):
            status, last_msg = "FINISHED", "市价指令剩余撤销"
        if order.get("time_condition") == "IOC":  # IOC 立即成交，限价下单且不能成交的价格，直接撤单
            if order["direction"] == "BUY" and price < ask_price or order["direction"] == "SELL" and price > bid_price:
                status, last_msg = "FINISHED", "已撤单报单已提交"
        if order["direction"] == "BUY" and price >= ask_price or order["direction"] == "SELL" and price <= bid_price:
            status, last_msg = "FINISHED", "全部成交"
        return status, last_msg, price



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'yanqiong'


import asyncio
from typing import Optional, Union

from tqsdk.algorithm.time_table_generater import _gen_random_list
from tqsdk.api import TqApi
from tqsdk.channel import TqChan
from tqsdk.datetime import _get_trading_timestamp, _get_trade_timestamp
from tqsdk.lib import InsertOrderUntilAllTradedTask
from tqsdk.tradeable import TqAccount, TqKq, TqSim
from tqsdk.rangeset import _rangeset_slice, _rangeset_head, _rangeset_length


class Twap(object):
    """
    天勤算法 - Twap

    Twap 算法实现了在设定的交易时间段内，完成设定的下单手数。

    构造 Twap 类的实例，该算法实例就会开始运行，根据以下逻辑下单：

    1. 将用户设置的总手数，拆分为一个随机手数列表，列表的值即为每次下单的手数，列表元素之和为总下单手数，同时每次下单手数也符合用户设置的每次下单手数的上下限；
    2. 将总的交易时间段拆分为随机时间间隔列表，列表的值即为每次下单的时间间隔，这些时间间隔相加应该等于总的下单时间；
    3. 每一次下单，在两个列表中分别取出下单手数、下单预计完成的时间，先用跟盘价下单，在当前时间间隔已经过去 2/3 或者只剩下 2s 时，主动撤掉未成交单，用对手价下单剩余手数；
    4. 在当前时间段已结束并且下单手数全部成交完，会开始下一次下单，重复第 3 步。

    基于以上逻辑，用户参数应该满足：

    平均每次下单时间 = duration / 下单次数 > 3s

    其中，下单次数 = 总的下单手数 / 平均每次下单手数 = 总的下单手数 / ((单次委托单最小下单手数 + 单次委托单最大下单手数) / 2)


    **注意**：

    时间段 duration，以 s 为单位，时长可以跨非交易时间段，但是不可以跨交易日。

    比如，SHFE.cu2101 的白盘交易时间段为 ["09:00:00" ～ "10:15:00"], ["10:30:00", "11:30:00"], ["13:30:00", "15:00:00"]，duration 设置为 1200 (20分钟)。

    如果当前行情时间是 2020-09-15 09:10:00，那么下单的时间应该在 2020-09-15 09:10:00 ～ 2020-09-15 09:30:00；
    如果当前行情时间是 2020-09-15 10:10:00，那么下单的时间应该在 2020-09-15 10:10:00 ～ 2020-09-15 10:15:00，以及 2020-09-15 10:30:00 ～ 2020-09-15 10:45:00。


    本模块不支持在回测中使用。
    """

    def __init__(self, api: TqApi, symbol: str, direction: str, offset: str, volume: int, duration: float,
                 min_volume_each_order: int, max_volume_each_order: int,
                 account: Optional[Union[TqAccount, TqKq, TqSim]] = None):
        """
        创建 Twap 实例

        Args:
            api (TqApi): TqApi实例，该task依托于指定api下单/撤单

            symbol (str): 拟下单的合约symbol, 格式为 交易所代码.合约代码,  例如 "SHFE.cu1801"

            direction (str): "BUY" 或 "SELL"

            offset (str): "OPEN", "CLOSE"，"CLOSETODAY"

            volume (int): 需要下单的总手数

            duration (int): 算法执行的时长，以秒为单位，时长可以跨非交易时间段，但是不可以跨交易日
            * 设置为 60*10, 可以是 10:10～10:15 + 10:30~10:35

            min_volume_each_order (int):单笔最小委托单，每笔委托单数默认在最小和最大值中产生

            max_volume_each_order (int):单笔最大委托单，每笔委托单数默认在最小和最大值中产生

            account (TqAccount/TqKq/TqSim): [可选]指定发送下单指令的账户实例, 多账户模式下，该参数必须指定

        Example1::

          from tqsdk import TqApi
          from tqsdk.algorithm import Twap

          api = TqApi(auth="快期账户,用户密码")
          # 设置twap任务参数
          target_twap = Twap(api,"SHFE.rb2012","BUY","OPEN",500,300,10,25)
          # 启动循环
          while True:
            api.wait_update()
            if target_twap.is_finished():
                break
          api.close()

        Example2::

          from tqsdk import TqApi
          from tqsdk.algorithm import Twap

          api = TqApi(auth="快期账户,用户密码")
          target_twap = Twap(api,"SHFE.rb2012","BUY","OPEN",500,300,10,25)

          num_of_trades = 0

          while True:
            api.wait_update()

            if num_of_trades < len(target_twap.trades):
              # 最新的成交
              for i in range(num_of_trades - len(target_twap.trades), 0):
                print("新的成交", target_twap.trades[i])
              print(target_twap.average_trade_price)  # 打印出当前已经成交的平均价格
              num_of_trades = len(target_twap.trades)

            if target_twap.is_finished():
                break

          print("打印出 twap 全部成交以及成交均价")
          print(target_twap.trades)
          print(target_twap.average_trade_price)
          api.close()
        """
        if symbol.startswith("CZCE.ZC"):
            raise Exception("动力煤期货不支持创建 targetpostask、twap、vwap 任务，交易所规定该品种最小开仓手数为大于等于 4 手，这些函数还未支持该规则!")
        if symbol.startswith("CZCE.WH"):
            raise Exception("强麦期货不支持创建 targetpostask、twap、vwap 任务，交易所规定该品种最小开仓手数为大于等于 10 手，这些函数还未支持该规则!")
        if symbol.startswith("CZCE.PM"):
            raise Exception("普麦期货不支持创建 targetpostask、twap、vwap 任务，交易所规定该品种最小开仓手数为大于等于 10 手，这些函数还未支持该规则!")
        if symbol.startswith("CZCE.RI"):
            raise Exception("早籼稻期货不支持创建 targetpostask、twap、vwap 任务，交易所规定该品种最小开仓手数为大于等于 10 手，这些函数还未支持该规则!")
        if symbol.startswith("CZCE.JR"):
            raise Exception("粳稻期货不支持创建 targetpostask、twap、vwap 任务，交易所规定该品种最小开仓手数为大于等于 10 手，这些函数还未支持该规则!")
        if symbol.startswith("CZCE.LR"):
            raise Exception("晚籼稻期货不支持创建 targetpostask、twap、vwap 任务，交易所规定该品种最小开仓手数为大于等于 10 手，这些函数还未支持该规则!")
        if symbol == "CZCE.SA309" or symbol == "CZCE.SA310":
            raise Exception("纯碱期货 2309 合约及 2310 合约不支持创建 targetpostask、twap、vwap 任务，交易所规定该品种最小开仓手数为大于等于 4 手，这些函数还未支持该规则!")
        self._api = api
        self._account = api._account._check_valid(account)
        if self._account is None:
            raise Exception(f"多账户模式下, 需要指定账户实例 account")
        self._symbol = symbol
        self._direction = direction
        self._offset = offset
        self._volume = int(volume)
        self._duration = duration
        self._min_volume_each_order = int(min_volume_each_order)
        self._max_volume_each_order = int(max_volume_each_order)
        if self._max_volume_each_order <= 0 or self._min_volume_each_order <= 0:
            raise Exception("请调整参数, min_volume_each_order、max_volume_each_order 必须是大于 0 的整数。")
        if self._min_volume_each_order > self._max_volume_each_order:
            raise Exception("请调整参数, min_volume_each_order 必须小于 max_volume_each_order。")
        # 得到有效的手数序列和时间间隔序列
        volume_list, interval_list = self._get_volume_list()
        self._task = self._api.create_task(self._run(volume_list, interval_list))
        self._order_task = None

        self.trades = []  # 所有的 trade 列表
        self._trade_sum_volume = 0  # 所有 trade 的成交的总手数
        self._trade_sum_amount = 0  # 所有 trade 的成交的总支出 （手数*价格）
        self._trade_objs_chan = TqChan(self._api)
        self._trade_recv_task = self._api.create_task(self._trade_recv())

    @property
    def average_trade_price(self):
        # 平均成交价格
        if self._trade_sum_volume == 0:
            return float('nan')
        else:
            return self._trade_sum_amount / self._trade_sum_volume

    async def _run(self, volume_list, interval_list):
        self._quote = await self._api.get_quote(self._symbol)
        # 计算得到时间序列，每个时间段快要结束的时间点，此时应该从被动价格切换为主动价格
        deadline_timestamp_list, strict_deadline_timestamp_list = self._get_deadline_timestamp(interval_list)
        for i in range(len(volume_list)):
            exit_immediately = (i + 1 == len(volume_list))  # 如果是最后一个时间段，需要全部成交后立即退出
            await self._insert_order(volume_list[i], deadline_timestamp_list[i], strict_deadline_timestamp_list[i],
                                     exit_immediately)

    async def _trade_recv(self):
        try:
            async for trade in self._trade_objs_chan:
                self.trades.append(trade)
                self._trade_sum_volume += trade['volume']
                self._trade_sum_amount += trade['volume'] * trade['price']
        finally:
            await self._trade_objs_chan.close()

    def _get_deadline_timestamp(self, interval_list):
        # interval - min(5, interval / 3)  # 定义一个时间片段中，开始到快要结束的时间间隔
        # 使用 rangeSet 计算出时间段；都使用本地时间或者都使用 **行情时间**
        # 当前交易日完整的交易时间段
        trading_timestamp = _get_trading_timestamp(self._quote, self._quote.datetime)
        trading_timestamp_nano_range = trading_timestamp['night'] + trading_timestamp['day']  # 当前交易日完整的交易时间段
        # 当前时间 行情时间
        current_timestamp_nano = _get_trade_timestamp(self._quote.datetime, float('nan'))
        if not trading_timestamp_nano_range[0][0] <= current_timestamp_nano < trading_timestamp_nano_range[-1][1]:
            raise Exception("当前时间不在指定的交易时间段内")
        # 此时，current_timestamp_nano 一定在此交易日内
        deadline_timestamp_list = []
        strict_deadline_timestamp_list = []
        for interval in interval_list:
            r = _rangeset_head(_rangeset_slice(trading_timestamp_nano_range, current_timestamp_nano), int(interval*1e9))
            strict_interval = interval - min(2, interval / 3)
            strict_r = _rangeset_head(_rangeset_slice(trading_timestamp_nano_range, current_timestamp_nano), int(strict_interval*1e9))
            if _rangeset_length(r) < int(interval*1e9):
                raise Exception("指定时间段超出当前交易日")
            deadline_timestamp_list.append(r[-1][1])
            strict_deadline_timestamp_list.append(strict_r[-1][1])
            current_timestamp_nano = r[-1][1]
        return deadline_timestamp_list, strict_deadline_timestamp_list

    async def _insert_order(self, volume, end_time, strict_end_time, exit_immediately):
        volume_left = volume
        try:
            trade_chan = TqChan(self._api)
            self._order_task = InsertOrderUntilAllTradedTask(self._api, self._symbol, self._direction, self._offset,
                                                             volume=volume, price="PASSIVE", trade_chan=trade_chan,
                                                             trade_objs_chan=self._trade_objs_chan,
                                                             account=self._account)
            async with self._api.register_update_notify() as update_chan:
                async for _ in update_chan:
                    if _get_trade_timestamp(self._quote.datetime, float('nan')) > strict_end_time:
                        break
                    else:
                        while not trade_chan.empty():
                            v = await trade_chan.recv()
                            volume_left = volume_left - (v if self._direction == "BUY" else -v)
                        if exit_immediately and volume_left == 0:
                            break
        finally:
            self._order_task._task.cancel()
            await asyncio.gather(self._order_task._task, return_exceptions=True)
            while not trade_chan.empty():
                v = await trade_chan.recv()
                volume_left = volume_left - (v if self._direction == "BUY" else -v)
            await trade_chan.close()
            if volume_left > 0:
                await self._insert_order_active(volume_left)

    async def _insert_order_active(self, volume):
        try:
            trade_chan = TqChan(self._api)
            self._order_task = InsertOrderUntilAllTradedTask(self._api, self._symbol, self._direction, self._offset,
                                                             volume=volume, price="ACTIVE", trade_chan=trade_chan,
                                                             trade_objs_chan=self._trade_objs_chan,
                                                             account=self._account)
            async for v in trade_chan:
                volume = volume - (v if self._direction == "BUY" else -v)
                if volume == 0:
                    break
        finally:
            await trade_chan.close()
            self._order_task._task.cancel()
            await asyncio.gather(self._order_task._task, return_exceptions=True)

    def _get_volume_list(self):
        if self._volume < self._max_volume_each_order:
            return [self._volume], [self._duration]
        # 先确定 volume_list
        volume_list = _gen_random_list(sum_val=self._volume, min_val=self._min_volume_each_order, max_val=self._max_volume_each_order)
        interval = int(self._duration / len(volume_list))
        if interval < 3:
            raise Exception("请调整参数, 每次下单时间间隔不能小于3s, 将单次下单手数阈值调大或者增长下单时间。")
        min_interval = int(max(3, interval - 2))
        max_interval = int(interval * 2 - max(3, interval - 2)) + 1
        interval_list = _gen_random_list(sum_val=self._duration, min_val=min_interval, max_val=max_interval, length=len(volume_list))
        return volume_list, interval_list

    def cancel(self):
        """
        取消当前 Twap 算法实例，会将该实例已经发出但还是未成交的委托单撤单。

        Example::

          from tqsdk import TqApi
          from tqsdk.algorithm import Twap

          api = TqApi(auth="快期账户,用户密码")
          # 设置twap任务参数
          quote = api.get_quote("SHFE.rb2012")
          target_twap = Twap(api,"SHFE.rb2012","BUY","OPEN",500,300,10,25)
          api.wait_update()
          # 运行代码。。。
          target_twap.cancel()
          while True:
            api.wait_update()
            if target_twap.is_finished():
                break
          api.close()
        """
        if self._task.done():
            return
        self._task.cancel()

    def is_finished(self):
        """
        返回当前 Twap 算法实例是否已经结束。即此实例不会再发出下单或者撤单的任何动作。

        Returns:
            bool: 当前 Twap 算法实例是否已经结束
        """
        return self._task.done()



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'yangyang'

import asyncio
import csv
import os
from datetime import date, datetime
from encodings.utf_8 import StreamWriter
from typing import Union, List, Optional
import lzma

import pandas

from tqsdk.api import TqApi
from tqsdk.channel import TqChan
from tqsdk.datetime import _get_trading_day_start_time, _get_trading_day_end_time, _datetime_to_timestamp_nano
from tqsdk.diff import _get_obj
from tqsdk.tafunc import get_dividend_df, get_dividend_factor
from tqsdk.utils import _generate_uuid

try:
    file_path = os.path.split(os.path.abspath(__file__))[0]
    with lzma.open(os.path.join(file_path, "dead_ins.lzma"), "rt", encoding="utf-8") as dead_ins_file:
        DEAD_INS = {l.strip() for l in dead_ins_file}
except:
    DEAD_INS = {}


# 价格相关的字段，需要 format 数据格式
PRICE_KEYS = ["open", "high", "low", "close", "last_price", "highest", "lowest"] + [f"bid_price{i}" for i in range(1, 6)] + [f"ask_price{i}" for i in range(1, 6)]


class DataDownloader:
    """
    数据下载工具是 TqSdk 专业版中的功能，能让用户下载目前 TqSdk 提供的全部期货、期权和股票类的历史数据，下载数据支持 tick 级别精度和任意 kline 周期

    如果想使用数据下载工具，可以点击 `天勤量化专业版 <https://www.shinnytech.com/tqsdk_professional/>`_ 申请使用或购买

    历史数据下载器, 输出到csv文件

    多合约按时间横向对齐
    """

    def __init__(self, api: TqApi, symbol_list: Union[str, List[str]], dur_sec: int, start_dt: Union[date, datetime],
                 end_dt: Union[date, datetime], csv_file_name: Union[str, asyncio.StreamWriter],
                 adj_type: Union[str, None] = None) -> None:
        """
        创建历史数据下载器实例

        Args:
            api (TqApi): TqApi实例，该下载器将使用指定的api下载数据

            symbol_list (str/list of str): 需要下载数据的合约代码，当指定多个合约代码时将其他合约按第一个合约的交易时间对齐

            dur_sec (int): 数据周期，以秒为单位。例如: 1分钟线为60,1小时线为3600,日线为86400,Tick数据为0

            start_dt (date/datetime): 起始时间, 如果类型为 date 则指的是交易日, 如果为 datetime 则指的是具体时间点

            end_dt (date/datetime): 结束时间, 如果类型为 date 则指的是交易日, 如果为 datetime 则指的是具体时间点

            csv_file_name (str/StreamWriter): [必填]输出方式:
                * str : 输出 csv 的文件名

                * StreamWriter: 直接将内容输出到 StreamWriter

            adj_type (str/None): 复权计算方式，默认值为 None。"F" 为前复权；"B" 为后复权；None 表示不复权。只对股票、基金合约有效。

        Example::

            from datetime import datetime, date
            from contextlib import closing
            from tqsdk import TqApi, TqAuth, TqSim
            from tqsdk.tools import DataDownloader

            api = TqApi(auth=TqAuth("快期账户", "账户密码"))
            download_tasks = {}
            # 下载从 2018-01-01 到 2018-09-01 的 SR901 日线数据
            download_tasks["SR_daily"] = DataDownloader(api, symbol_list="CZCE.SR901", dur_sec=24*60*60,
                                start_dt=date(2018, 1, 1), end_dt=date(2018, 9, 1), csv_file_name="SR901_daily.csv")
            # 下载从 2017-01-01 到 2018-09-01 的 rb主连 5分钟线数据
            download_tasks["rb_5min"] = DataDownloader(api, symbol_list="KQ.m@SHFE.rb", dur_sec=5*60,
                                start_dt=date(2017, 1, 1), end_dt=date(2018, 9, 1), csv_file_name="rb_5min.csv")
            # 下载从 2018-01-01凌晨6点 到 2018-06-01下午4点 的 cu1805,cu1807,IC1803 分钟线数据，所有数据按 cu1805 的时间对齐
            # 例如 cu1805 夜盘交易时段, IC1803 的各项数据为 N/A
            # 例如 cu1805 13:00-13:30 不交易, 因此 IC1803 在 13:00-13:30 之间的K线数据会被跳过
            download_tasks["cu_min"] = DataDownloader(api, symbol_list=["SHFE.cu1805", "SHFE.cu1807", "CFFEX.IC1803"], dur_sec=60,
                                start_dt=datetime(2018, 1, 1, 6, 0 ,0), end_dt=datetime(2018, 6, 1, 16, 0, 0), csv_file_name="cu_min.csv")
            # 下载从 2018-05-01凌晨0点 到 2018-06-01凌晨0点 的 T1809 盘口Tick数据
            download_tasks["T_tick"] = DataDownloader(api, symbol_list=["CFFEX.T1809"], dur_sec=0,
                                start_dt=datetime(2018, 5, 1), end_dt=datetime(2018, 6, 1), csv_file_name="T1809_tick.csv")
            # 使用with closing机制确保下载完成后释放对应的资源
            with closing(api):
                while not all([v.is_finished() for v in download_tasks.values()]):
                    api.wait_update()
                    print("progress: ", { k:("%.2f%%" % v.get_progress()) for k,v in download_tasks.items() })
        """
        self._api = api
        if not self._api._auth._has_feature("tq_dl"):
            raise Exception("您的账户不支持下载历史数据功能，需要购买后才能使用。升级网址：https://www.shinnytech.com/tqsdk_professional/")
        if isinstance(start_dt, datetime):
            self._start_dt_nano = _datetime_to_timestamp_nano(start_dt)
        else:
            self._start_dt_nano = _get_trading_day_start_time(_datetime_to_timestamp_nano(datetime(start_dt.year, start_dt.month, start_dt.day)))
        if isinstance(end_dt, datetime):
            self._end_dt_nano = _datetime_to_timestamp_nano(end_dt)
        else:
            self._end_dt_nano = _get_trading_day_end_time(_datetime_to_timestamp_nano(datetime(end_dt.year, end_dt.month, end_dt.day)))
        self._current_dt_nano = self._start_dt_nano
        self._symbol_list = symbol_list if isinstance(symbol_list, list) else [symbol_list]
        # 下载合约超时时间（默认 30s），已下市的没有交易的合约，超时时间可以设置短一点（2s），用户不希望自己的程序因为没有下载到数据而中断
        self._timeout_seconds = 2 if any([symbol in DEAD_INS for symbol in self._symbol_list]) else 30
        self._dur_nano = dur_sec * 1000000000
        if self._dur_nano == 0 and len(self._symbol_list) != 1:
            raise Exception("Tick序列不支持多合约")
        if adj_type not in [None, "F", "B", "FORWARD", "BACK"]:
            raise Exception("adj_type 参数只支持 None (不复权) ｜ 'F' (前复权) ｜ 'B' (后复权)")
        self._adj_type = adj_type[0] if adj_type else adj_type
        if isinstance(csv_file_name, str) or isinstance(csv_file_name, asyncio.StreamWriter):
            self._csv_file_name = csv_file_name
        else:
            raise Exception("csv_file_name 参数只支持 str ｜ StreamWriter 类型")
        self._csv_header = self._get_headers()
        # 缓存合约对应的复权系数矩阵，每个合约只计算一次
        # 含义为截止 datetime 之前(不包含) 应使用 factor 复权
        self._dividend_cache = {}
        self._data_series = None
        self._task = self._api.create_task(self._run())

    def is_finished(self) -> bool:
        """
        判断是否下载完成

        Returns:
            bool: 如果数据下载完成则返回 True, 否则返回 False.
        """
        return self._task.done()

    def get_progress(self) -> float:
        """
        获得下载进度百分比

        Returns:
            float: 下载进度,100表示下载完成
        """
        return 100.0 if self._task.done() else (self._current_dt_nano - self._start_dt_nano) / (
                self._end_dt_nano - self._start_dt_nano) * 100

    def _get_data_series(self) -> Optional[pandas.DataFrame]:
        """
        获取下载的 DataFrame 格式数据

        todo: 在 utils 中增加工具函数，返回与 kline 一致的数据结构

        Returns:
            pandas.DataFrame/None: 下载的 klines 或者 ticks 数据，DataFrame 格式。下载完成前返回 None。


        Example::

            from datetime import datetime, date
            rom tqsdk import TqApi, TqAuth
            from contextlib import closing
            from tqsdk.tools import DataDownloader

            api = TqApi(auth=TqAuth("快期账户", "账户密码"))
            # 下载从 2018-06-01 到 2018-09-01 的 SR901 日线数据
            download_task = DataDownloader(api, symbol_list="CZCE.SR901", dur_sec=24*60*60,
                                start_dt=date(2018, 6, 1), end_dt=date(2018, 9, 1), csv_file_name="klines.csv")
            # 使用with closing机制确保下载完成后释放对应的资源
            with closing(api):
                while not download_task.is_finished():
                    api.wait_update()
                    print(f"progress: {download_task.get_progress():.2} %")
                print(download_task._get_data_series())
        """
        if not self._task.done():
            return None
        if isinstance(self._csv_file_name, str):
            if not self._data_series:
                self._data_series = pandas.read_csv(self._csv_file_name)
            return self._data_series
        else:
            raise Exception('DataDownloader._get_data_series 接口仅支持 csv_file_name 參數为 str 时使用')

    async def _ensure_dividend_factor(self, quote, timestamp):
        if quote.instrument_id not in self._dividend_cache:
            # 对每个除权除息矩阵增加 factor 序列，为当日的复权因子
            df = get_dividend_df(quote.stock_dividend_ratio, quote.cash_dividend_ratio)
            df = df[df["datetime"].between(timestamp, self._end_dt_nano, inclusive="right")]  # 只需要第一笔行情时间～结束时间之间的复权因子, 左开右闭
            df["pre_close"] = float('nan')  # 初始化 pre_close 为 nan
            for i in range(len(df)):
                chart_info = {
                    "aid": "set_chart",
                    "chart_id": _generate_uuid("PYSDK_downloader"),
                    "ins_list": quote.instrument_id,
                    "duration": 86400 * 1000000000,
                    "view_width": 2,
                    "focus_datetime": int(df["datetime"].iloc[i]),
                    "focus_position": 1
                }
                await self._api._send_chan.send(chart_info)
                chart = _get_obj(self._api._data, ["charts", chart_info["chart_id"]])
                serial = _get_obj(self._api._data, ["klines", quote.instrument_id, str(86400000000000)])
                try:
                    async with self._api.register_update_notify() as update_chan:
                        async for _ in update_chan:
                            if not (chart_info.items() <= _get_obj(chart, ["state"]).items()):
                                continue  # 当前请求还没收齐回应, 不应继续处理
                            left_id = chart.get("left_id", -1)
                            right_id = chart.get("right_id", -1)
                            if (left_id == -1 and right_id == -1) or self._api._data.get("mdhis_more_data", True) or serial.get("last_id", -1) == -1:
                                continue  # 定位信息还没收到, 或数据序列还没收到, 合约的数据是否收到
                            last_item = serial["data"].get(str(left_id), {})
                            # 复权时间点的昨收盘
                            df.loc[i, 'pre_close'] = last_item['close'] if last_item.get('close') else float('nan')
                            break
                finally:
                    await self._api._send_chan.send({
                        "aid": "set_chart",
                        "chart_id": chart_info["chart_id"],
                        "ins_list": "",
                        "duration": 86400000000000,
                        "view_width": 2
                    })
            df["factor"] = (df["pre_close"] - df["cash_dividend"]) / df["pre_close"] / (1 + df["stock_dividend"])
            df["factor"].fillna(1.0, inplace=True)
            # 插入结束时间这条记录, 因为可能存在行情时间等于 _end_dt_nano 的行情，因此这里 +1
            df = df.append({"datetime": self._end_dt_nano+1, "factor": 1.0}, ignore_index=True)
            if self._adj_type == "F":
                df["factor"] = df["factor"].iloc[::-1].cumprod().iloc[::-1]
            elif self._adj_type == "B":
                # 后复权按定义上应该从第一笔行情之后产生的复权事件开始
                # 第一笔行情时间一定小于 df["datetime"].iloc[0], 因此复权是从 df["datetime"].iloc[0] 开始
                df["factor"] = 1.0 / df["factor"].cumprod()
                # 至此 df 每行的含义为从 datetime 开始应使用 factor 复权
                # 该格式并不好用，需要改为截止 datetime 之前(不包含) 应使用 factor 复权
                df["factor"] = df["factor"].shift(1)
                df["factor"].iloc[0] = 1.0
            self._dividend_cache[quote.instrument_id] = {
                "df": df,
                "last_dt": 0,
                "factor": float("nan"),
            }

    async def _run(self):
        self._quote_list = await self._api.get_quote_list(self._symbol_list)
        # 下载数据的 async generator
        gen = self._download_data()
        # cols 是复权需要重新计算的列名
        index_datetime_nano = self._csv_header.index("datetime_nano")
        if self._dur_nano != 0:
            cols = ["open", "high", "low", "close"]
        else:
            cols = ["last_price", "highest", "lowest"]
            cols.extend(f"{x}{i}" for x in ["bid_price", "ask_price"] for i in range(1, 6))
        try:
            if isinstance(self._csv_file_name, asyncio.StreamWriter):
                writer = StreamWriter(self._csv_file_name)
            else:
                writer = open(self._csv_file_name, 'w', newline='')
            csv_writer = csv.writer(writer, dialect='excel')
            csv_writer.writerow(self._csv_header)
            async for item in gen:
                for quote in self._quote_list:
                    symbol = quote.instrument_id
                    if self._adj_type and quote.ins_class in ["STOCK", "FUND"]:
                        # 如果存在 STOCK / FUND 并且 adj_type is not None, 这里需要提前准备下载时间段内的复权因子
                        # 前复权需要提前计算除权因子
                        await self._ensure_dividend_factor(quote, item[index_datetime_nano])
                        dividend_cache = self._dividend_cache[symbol]
                        # dividend_df 和 _data_chan 中取出的数据都是按时间升序排列的，因此可以使用归并算法
                        if dividend_cache["last_dt"] <= item[index_datetime_nano]:
                            dividend_df = dividend_cache["df"]
                            dividend_df = dividend_df[dividend_df["datetime"].gt(item[index_datetime_nano])]
                            dividend_cache["df"] = dividend_df
                            dividend_cache["last_dt"] = dividend_df["datetime"].iloc[0]
                            dividend_cache["factor"] = dividend_df["factor"].iloc[0]
                        if dividend_cache["factor"] != 1:
                            item = item.copy()
                            for c in cols:  # datetime_nano
                                index = self._csv_header.index(f"{symbol}.{c}")
                                item[index] = item[index] * dividend_cache["factor"]
                csv_writer.writerow(item)
                if isinstance(self._csv_file_name, asyncio.StreamWriter):
                    await self._csv_file_name.drain()
        finally:
            if isinstance(self._csv_file_name, asyncio.StreamWriter):
                self._csv_file_name.write_eof()
            else:
                writer.close()
            # 这里 `await gen.aclose()` 实际测试代码与文档描述不符，无论文件全部下载完正常退出还是写文件过程中抛出例外退出，都没有再抛出任何例外
            # https://docs.python.org/3/reference/expressions.html#agen.aclose
            await gen.aclose()

    async def _timeout_handle(self, timeout, chart):
        await asyncio.sleep(timeout)
        if chart.get("left_id", -1) == -1 and chart.get("right_id", -1) == -1:
            self._task.cancel()

    async def _download_data(self):
        """下载数据, 多合约横向按时间对齐"""
        chart_info = {
            "aid": "set_chart",
            "chart_id": _generate_uuid("PYSDK_downloader"),
            "ins_list": ",".join(self._symbol_list),
            "duration": self._dur_nano,
            "view_width": 2000,
            "focus_datetime": self._start_dt_nano,
            "focus_position": 0,
        }
        # 还没有发送过任何请求, 先请求定位左端点
        await self._api._send_chan.send(chart_info)
        chart = _get_obj(self._api._data, ["charts", chart_info["chart_id"]])
        # 增加一个 task，在 30s 后检查 chart 是否返回了左右 id 范围，如果没有就 cancel self._task，防止程序一直卡在那里
        timeout_task = self._api.create_task(self._timeout_handle(self._timeout_seconds, chart))
        current_id = None  # 当前数据指针
        data_cols = self._get_data_cols()
        serials = []
        for symbol in self._symbol_list:
            path = ["klines", symbol, str(self._dur_nano)] if self._dur_nano != 0 else ["ticks", symbol]
            serial = _get_obj(self._api._data, path)
            serials.append(serial)
        try:
            async with self._api.register_update_notify() as update_chan:
                async for _ in update_chan:
                    if not (chart_info.items() <= _get_obj(chart, ["state"]).items()):
                        # 当前请求还没收齐回应, 不应继续处理
                        continue
                    left_id = chart.get("left_id", -1)
                    right_id = chart.get("right_id", -1)
                    if (left_id == -1 and right_id == -1) or chart.get("more_data", True):
                        # 定位信息还没收到, 或数据序列还没收到
                        continue
                    # 检查合约的数据是否收到
                    if any([serial.get("last_id", -1) == -1 for serial in serials]):
                        continue
                    if current_id is None:
                        current_id = max(left_id, 0)
                    while current_id <= right_id:
                        item = serials[0]["data"].get(str(current_id), {})
                        if item.get("datetime", 0) == 0 or item["datetime"] > self._end_dt_nano:
                            # 当前 id 已超出 last_id 或k线数据的时间已经超过用户限定的右端
                            return
                        row = [self._nano_to_str(item["datetime"]), item["datetime"]]
                        for col in data_cols:
                            row.append(self._get_value(item, col, self._quote_list[0]["price_decs"]))
                        for i in range(1, len(self._symbol_list)):
                            symbol = self._symbol_list[i]
                            tid = serials[0].get("binding", {}).get(symbol, {}).get(str(current_id), -1)
                            k = {} if tid == -1 else serials[i]["data"].get(str(tid), {})
                            for col in data_cols:
                                row.append(self._get_value(k, col, self._quote_list[i]["price_decs"]))
                        yield row
                        current_id += 1
                        self._current_dt_nano = item["datetime"]
                    # 当前 id 已超出订阅范围, 需重新订阅后续数据
                    chart_info.pop("focus_datetime", None)
                    chart_info.pop("focus_position", None)
                    chart_info["left_kline_id"] = current_id
                    await self._api._send_chan.send(chart_info)
        finally:
            # 释放chart资源
            await self._api._send_chan.send({
                "aid": "set_chart",
                "chart_id": chart_info["chart_id"],
                "ins_list": "",
                "duration": self._dur_nano,
                "view_width": 2000,
            })
            timeout_task.cancel()
            await asyncio.gather(timeout_task, return_exceptions=True)

    def _get_headers(self):
        data_cols = self._get_data_cols()
        return ["datetime", "datetime_nano"] + [f"{symbol}.{col}" for symbol in self._symbol_list for col in data_cols]

    def _get_data_cols(self):
        if self._dur_nano != 0:
            return ["open", "high", "low", "close", "volume", "open_oi", "close_oi"]
        else:
            cols = ["last_price", "highest", "lowest", "average", "volume", "amount", "open_interest"]
            price_range = 1
            for symbol in self._symbol_list:
                if symbol.split('.')[0] in {"SHFE", "INE", "SSE", "SZSE"}:
                    price_range = 5
                    break
            for i in range(price_range):
                cols.extend(f"{x}{i+1}" for x in ["bid_price", "bid_volume", "ask_price", "ask_volume"])
            return cols

    @staticmethod
    def _get_value(obj, key, price_decs):
        try:
            if key in PRICE_KEYS:
                return round(obj[key], price_decs)
            else:
                return obj[key]
        except KeyError:
            return "#N/A"
        except TypeError:
            return float("nan")

    @staticmethod
    def _nano_to_str(nano):
        dt = datetime.fromtimestamp(nano // 1000000000)
        return "%d-%02d-%02d %02d:%02d:%02d.%09d" % (dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, int(nano) % 1000000000)
